import copy
from contextlib import contextmanager

import pytest

import weave
from weave.legacy.weave import graph, storage
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.panels.panel_plot import Plot, PlotConstants, Series

from .test_run_segment import create_experiment


@contextmanager
def const_nodes_equal():
    graph.ConstNode.__eq__ = (
        lambda self, other: isinstance(other, graph.ConstNode)
        and self.val == other.val
        and self.type == other.type
    )
    yield
    del graph.ConstNode.__eq__


def test_run_segment_plot_config():
    last_segment = create_experiment(1000, 3, 0.8)
    weave.save(last_segment)
    config = Plot(last_segment.experiment()).config
    assert len(config.series) == 1
    assert all(
        isinstance(v, (graph.VoidNode, graph.ConstNode))
        for v in config.series[0].table.columnSelectFunctions.values()
    )


def test_multi_series_plot_config_with_grouping():
    last_segment = create_experiment(1000, 3, 0.8)
    weave.save(last_segment)
    plot = Plot(last_segment.experiment())
    plot.set_x(
        lambda row: weave.legacy.weave.ops.number_bin(
            row["step"], weave.legacy.weave.ops.numbers_bins_equal([1, 2000], 2)
        )
    )
    plot.set_y(lambda row: weave.legacy.weave.ops.numbers_avg(row["metric0"]))

    plot.groupby_x()
    plot.set_mark_constant("line")

    series2 = plot.config.series[0].clone()
    plot.add_series(series2)

    series2.set_y(lambda row: weave.legacy.weave.ops.numbers_min(row["metric0"]))
    series2.set_y2(lambda row: weave.legacy.weave.ops.numbers_max(row["metric0"]))
    series2.set_mark_constant("area")

    plot.groupby_x()

    assert len(plot.config.series) == 2
    assert all(
        isinstance(v, (graph.VoidNode, graph.ConstNode, graph.OutputNode))
        for v in list(plot.config.series[0].table.columnSelectFunctions.values())
        + list(plot.config.series[1].table.columnSelectFunctions.values())
    )


def test_multi_series_grouping():
    last_segment = create_experiment(1000, 3, 0.8)
    weave.save(last_segment)
    plot = Plot(last_segment.experiment())
    plot.set_x(
        lambda row: weave.legacy.weave.ops.number_bin(
            row["step"], weave.legacy.weave.ops.numbers_bins_equal([1, 2000], 2)
        )
    )
    plot.groupby_x()

    plot.set_y(lambda row: weave.legacy.weave.ops.numbers_avg(row["metric0"]))
    plot.set_mark_constant("line")

    series2 = plot.config.series[0].clone()
    plot.add_series(series2)

    series2.set_y(lambda row: weave.legacy.weave.ops.numbers_min(row["metric0"]))
    series2.set_y2(lambda row: weave.legacy.weave.ops.numbers_max(row["metric0"]))
    series2.set_mark_constant("area")

    plot2 = copy.deepcopy(plot)

    # compare plot-level grouping to series by series grouping
    for series in plot2.config.series:
        series.groupby_x()

    with const_nodes_equal():
        assert plot.config == plot2.config


def test_multi_series_setting():
    last_segment = create_experiment(1000, 3, 0.8)
    weave.save(last_segment)
    plot = Plot(last_segment.experiment())
    plot.set_x(
        lambda row: weave.legacy.weave.ops.number_bin(
            row["step"], weave.legacy.weave.ops.numbers_bins_equal([1, 2000], 2)
        )
    )

    plot.set_y(lambda row: row["metric0"])
    plot.set_mark_constant("line")

    series2 = plot.config.series[0].clone()
    plot.add_series(series2)

    series2.set_y(lambda row: row["metric0"])
    series2.set_y2(lambda row: row["metric0"])
    series2.set_mark_constant("area")

    plot2 = copy.deepcopy(plot)

    # compare plot-level setting to series by series setting
    plot.set_x(lambda row: row["step"])
    for series in plot2.config.series:
        series.set_x(lambda row: row["step"])

    with const_nodes_equal():
        assert plot.config == plot2.config


def test_constructor():
    last_segment = create_experiment(1000, 3, 0.8)
    weave.save(last_segment)
    series = Series(
        input_node=last_segment.experiment(),
        select_functions={
            "x": lambda row: row["step"],
            "y": lambda row: row["metric0"],
        },
    )
    plot = Plot(series=series)
    plot2 = Plot(series=[series])

    with const_nodes_equal():
        assert plot.config == plot2.config
    assert series.constants == PlotConstants(
        label="series",
        pointShape="circle",
        lineStyle="solid",
    )


@pytest.mark.skip(
    "I have to constantly update these random values and it doesn't always agree with CI"
)
def test_actual_config_value(fixed_random_seed):
    last_segment = create_experiment(1000, 3, 0.8)
    weave.save(last_segment)
    plot = Plot(last_segment.experiment())
    plot.set_x(
        lambda row: weave.legacy.weave.ops.number_bin(
            row["step"], weave.legacy.weave.ops.numbers_bins_equal([1, 2000], 2)
        )
    )
    plot.set_y(lambda row: weave.legacy.weave.ops.numbers_avg(row["metric0"]))
    plot.set_mark_constant("line")

    series2 = plot.config.series[0].clone()
    plot.add_series(series2)

    series2.set_y(lambda row: weave.legacy.weave.ops.numbers_min(row["metric0"]))
    series2.set_y2(lambda row: weave.legacy.weave.ops.numbers_max(row["metric0"]))
    series2.set_mark_constant("area")
    assert storage.to_python(plot.config) == {
        "_type": {
            "type": "PlotConfig",
            "_is_object": True,
            "series": {
                "type": "list",
                "objectType": {
                    "type": "Series",
                    "_is_object": True,
                    "table": {
                        "type": "TableState",
                        "_is_object": True,
                        "columns": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "BPZ7OF3CU3BTPJ": {
                                    "type": "PanelDef",
                                    "_is_object": True,
                                    "panelConfig": "none",
                                    "panelId": "string",
                                },
                                "3TAMZ9RE8CSQNQ": {
                                    "type": "PanelDef",
                                    "_is_object": True,
                                    "panelConfig": "none",
                                    "panelId": "string",
                                },
                                "V4ZUZP68VEFUE4": {
                                    "type": "PanelDef",
                                    "_is_object": True,
                                    "panelConfig": "none",
                                    "panelId": "string",
                                },
                                "DX2S5BL68G31XO": {
                                    "type": "PanelDef",
                                    "_is_object": True,
                                    "panelConfig": "none",
                                    "panelId": "string",
                                },
                                "6YQY0NYGB338WC": {
                                    "type": "PanelDef",
                                    "_is_object": True,
                                    "panelConfig": "none",
                                    "panelId": "string",
                                },
                                "FSW0FWMRIMFYK6": {
                                    "type": "PanelDef",
                                    "_is_object": True,
                                    "panelConfig": "none",
                                    "panelId": "string",
                                },
                                "UYPJ7CMYZWWCGF": {
                                    "type": "PanelDef",
                                    "_is_object": True,
                                    "panelConfig": "none",
                                    "panelId": "string",
                                },
                                "L3IBEEOE6UHRRS": {
                                    "type": "PanelDef",
                                    "_is_object": True,
                                    "panelConfig": "none",
                                    "panelId": "string",
                                },
                            },
                        },
                        "input_node": {
                            "type": "function",
                            "inputTypes": {},
                            "outputType": "unknown",
                        },
                        "autoColumns": "boolean",
                        "preFilterFunction": {
                            "type": "function",
                            "inputTypes": {},
                            "outputType": "unknown",
                        },
                        "columnNames": {
                            "type": "dict",
                            "key_type": "string",
                            "objectType": "string",
                        },
                        "columnSelectFunctions": {
                            "type": "dict",
                            "key_type": "string",
                            "objectType": {
                                "type": "function",
                                "inputTypes": {},
                                "outputType": "unknown",
                            },
                        },
                        "order": {"type": "list", "objectType": "string"},
                        "groupBy": {"type": "list", "objectType": "string"},
                        "sort": {"type": "list", "objectType": "string"},
                        "pageSize": "int",
                        "page": "int",
                    },
                    "constants": {
                        "type": "PlotConstants",
                        "_is_object": True,
                        "mark": "string",
                        "pointShape": "none",
                        "lineStyle": "none",
                        "label": {
                            "type": "const",
                            "valType": "string",
                            "val": "series",
                        },
                    },
                    "uiState": {
                        "type": "PlotUIState",
                        "_is_object": True,
                        "pointShape": "string",
                        "label": "string",
                    },
                    "dims": {
                        "type": "DimConfig",
                        "_is_object": True,
                        "x": "string",
                        "y": "string",
                        "color": "string",
                        "label": "string",
                        "tooltip": "string",
                        "pointSize": "string",
                        "pointShape": "string",
                        "y2": "string",
                    },
                },
            },
            "axisSettings": {
                "type": "AxisSettings",
                "_is_object": True,
                "color": {
                    "type": "AxisSetting",
                    "_is_object": True,
                    "noLabels": "boolean",
                    "noTitle": "boolean",
                    "noTicks": "boolean",
                },
                "x": {
                    "type": "AxisSetting",
                    "_is_object": True,
                    "noLabels": "boolean",
                    "noTitle": "boolean",
                    "noTicks": "boolean",
                },
                "y": {
                    "type": "AxisSetting",
                    "_is_object": True,
                    "noLabels": "boolean",
                    "noTitle": "boolean",
                    "noTicks": "boolean",
                },
            },
            "legendSettings": {
                "type": "LegendSettings",
                "_is_object": True,
                "x": {
                    "type": "LegendSetting",
                    "_is_object": True,
                    "noLegend": "boolean",
                },
                "y": {
                    "type": "LegendSetting",
                    "_is_object": True,
                    "noLegend": "boolean",
                },
                "color": {
                    "type": "LegendSetting",
                    "_is_object": True,
                    "noLegend": "boolean",
                },
            },
            "configOptionsExpanded": {
                "type": "ConfigOptionsExpanded",
                "_is_object": True,
                "x": "boolean",
                "y": "boolean",
                "label": "boolean",
                "tooltip": "boolean",
                "mark": "boolean",
            },
            "configVersion": "int",
        },
        "_val": {
            "series": [
                {
                    "table": {
                        "columns": {
                            "BPZ7OF3CU3BTPJ": {"panelConfig": None, "panelId": ""},
                            "3TAMZ9RE8CSQNQ": {"panelConfig": None, "panelId": ""},
                            "V4ZUZP68VEFUE4": {"panelConfig": None, "panelId": ""},
                            "DX2S5BL68G31XO": {"panelConfig": None, "panelId": ""},
                            "6YQY0NYGB338WC": {"panelConfig": None, "panelId": ""},
                            "FSW0FWMRIMFYK6": {"panelConfig": None, "panelId": ""},
                            "UYPJ7CMYZWWCGF": {"panelConfig": None, "panelId": ""},
                            "L3IBEEOE6UHRRS": {"panelConfig": None, "panelId": ""},
                        },
                        "input_node": {
                            "nodeType": "output",
                            "type": {
                                "type": "ArrowWeaveList",
                                "objectType": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "step": "int",
                                        "string_col": "string",
                                        "metric0": "float",
                                        "metric1": "float",
                                        "metric2": "float",
                                        "metric3": "float",
                                        "metric4": "float",
                                        "metric5": "float",
                                        "metric6": "float",
                                        "metric7": "float",
                                        "metric8": "float",
                                        "metric9": "float",
                                        "metric10": "float",
                                        "metric11": "float",
                                        "metric12": "float",
                                        "metric13": "float",
                                        "metric14": "float",
                                        "metric15": "float",
                                        "metric16": "float",
                                        "metric17": "float",
                                        "metric18": "float",
                                        "metric19": "float",
                                        "metric20": "float",
                                        "metric21": "float",
                                        "metric22": "float",
                                        "metric23": "float",
                                        "metric24": "float",
                                        "metric25": "float",
                                        "metric26": "float",
                                        "metric27": "float",
                                        "metric28": "float",
                                        "metric29": "float",
                                        "metric30": "float",
                                        "metric31": "float",
                                        "metric32": "float",
                                        "metric33": "float",
                                        "metric34": "float",
                                        "metric35": "float",
                                        "metric36": "float",
                                        "metric37": "float",
                                        "metric38": "float",
                                        "metric39": "float",
                                        "metric40": "float",
                                        "metric41": "float",
                                        "metric42": "float",
                                        "metric43": "float",
                                        "metric44": "float",
                                        "metric45": "float",
                                        "metric46": "float",
                                        "metric47": "float",
                                        "metric48": "float",
                                        "metric49": "float",
                                        "metric50": "float",
                                        "metric51": "float",
                                        "metric52": "float",
                                        "metric53": "float",
                                        "metric54": "float",
                                        "metric55": "float",
                                        "metric56": "float",
                                        "metric57": "float",
                                        "metric58": "float",
                                        "metric59": "float",
                                        "metric60": "float",
                                        "metric61": "float",
                                        "metric62": "float",
                                        "metric63": "float",
                                        "metric64": "float",
                                        "metric65": "float",
                                        "metric66": "float",
                                        "metric67": "float",
                                        "metric68": "float",
                                        "metric69": "float",
                                        "metric70": "float",
                                        "metric71": "float",
                                        "metric72": "float",
                                        "metric73": "float",
                                        "metric74": "float",
                                        "metric75": "float",
                                        "metric76": "float",
                                        "metric77": "float",
                                        "metric78": "float",
                                        "metric79": "float",
                                        "metric80": "float",
                                        "metric81": "float",
                                        "metric82": "float",
                                        "metric83": "float",
                                        "metric84": "float",
                                        "metric85": "float",
                                        "metric86": "float",
                                        "metric87": "float",
                                        "metric88": "float",
                                        "metric89": "float",
                                        "metric90": "float",
                                        "metric91": "float",
                                        "metric92": "float",
                                        "metric93": "float",
                                        "metric94": "float",
                                        "metric95": "float",
                                        "metric96": "float",
                                        "metric97": "float",
                                        "metric98": "float",
                                        "run_name": "string",
                                    },
                                },
                            },
                            "fromOp": {
                                "name": "RunSegment-experiment",
                                "inputs": {
                                    "self": {
                                        "nodeType": "output",
                                        "type": {
                                            "type": "RunSegment",
                                            "_is_object": True,
                                            "prior_run_ref": "string",
                                            "prior_run_branch_index": "int",
                                            "metrics": {
                                                "type": "ArrowWeaveList",
                                                "objectType": {
                                                    "type": "typedDict",
                                                    "propertyTypes": {
                                                        "step": "int",
                                                        "string_col": "string",
                                                        "metric0": "float",
                                                        "metric1": "float",
                                                        "metric2": "float",
                                                        "metric3": "float",
                                                        "metric4": "float",
                                                        "metric5": "float",
                                                        "metric6": "float",
                                                        "metric7": "float",
                                                        "metric8": "float",
                                                        "metric9": "float",
                                                        "metric10": "float",
                                                        "metric11": "float",
                                                        "metric12": "float",
                                                        "metric13": "float",
                                                        "metric14": "float",
                                                        "metric15": "float",
                                                        "metric16": "float",
                                                        "metric17": "float",
                                                        "metric18": "float",
                                                        "metric19": "float",
                                                        "metric20": "float",
                                                        "metric21": "float",
                                                        "metric22": "float",
                                                        "metric23": "float",
                                                        "metric24": "float",
                                                        "metric25": "float",
                                                        "metric26": "float",
                                                        "metric27": "float",
                                                        "metric28": "float",
                                                        "metric29": "float",
                                                        "metric30": "float",
                                                        "metric31": "float",
                                                        "metric32": "float",
                                                        "metric33": "float",
                                                        "metric34": "float",
                                                        "metric35": "float",
                                                        "metric36": "float",
                                                        "metric37": "float",
                                                        "metric38": "float",
                                                        "metric39": "float",
                                                        "metric40": "float",
                                                        "metric41": "float",
                                                        "metric42": "float",
                                                        "metric43": "float",
                                                        "metric44": "float",
                                                        "metric45": "float",
                                                        "metric46": "float",
                                                        "metric47": "float",
                                                        "metric48": "float",
                                                        "metric49": "float",
                                                        "metric50": "float",
                                                        "metric51": "float",
                                                        "metric52": "float",
                                                        "metric53": "float",
                                                        "metric54": "float",
                                                        "metric55": "float",
                                                        "metric56": "float",
                                                        "metric57": "float",
                                                        "metric58": "float",
                                                        "metric59": "float",
                                                        "metric60": "float",
                                                        "metric61": "float",
                                                        "metric62": "float",
                                                        "metric63": "float",
                                                        "metric64": "float",
                                                        "metric65": "float",
                                                        "metric66": "float",
                                                        "metric67": "float",
                                                        "metric68": "float",
                                                        "metric69": "float",
                                                        "metric70": "float",
                                                        "metric71": "float",
                                                        "metric72": "float",
                                                        "metric73": "float",
                                                        "metric74": "float",
                                                        "metric75": "float",
                                                        "metric76": "float",
                                                        "metric77": "float",
                                                        "metric78": "float",
                                                        "metric79": "float",
                                                        "metric80": "float",
                                                        "metric81": "float",
                                                        "metric82": "float",
                                                        "metric83": "float",
                                                        "metric84": "float",
                                                        "metric85": "float",
                                                        "metric86": "float",
                                                        "metric87": "float",
                                                        "metric88": "float",
                                                        "metric89": "float",
                                                        "metric90": "float",
                                                        "metric91": "float",
                                                        "metric92": "float",
                                                        "metric93": "float",
                                                        "metric94": "float",
                                                        "metric95": "float",
                                                        "metric96": "float",
                                                        "metric97": "float",
                                                        "metric98": "float",
                                                    },
                                                },
                                            },
                                            "run_name": "string",
                                        },
                                        "fromOp": {
                                            "name": "get",
                                            "inputs": {
                                                "uri": {
                                                    "nodeType": "const",
                                                    "type": "string",
                                                    "val": "local-artifact:///tmp/weave/pytest/weave/tests/test_plot.py::test_actual_config_value (setup)/RunSegment/164b91e6b2832f79254c6da17e094563",
                                                }
                                            },
                                        },
                                    }
                                },
                            },
                        },
                        "autoColumns": False,
                        "preFilterFunction": {"nodeType": "void", "type": "invalid"},
                        "columnNames": {
                            "BPZ7OF3CU3BTPJ": "",
                            "3TAMZ9RE8CSQNQ": "",
                            "V4ZUZP68VEFUE4": "",
                            "DX2S5BL68G31XO": "",
                            "6YQY0NYGB338WC": "",
                            "FSW0FWMRIMFYK6": "",
                            "UYPJ7CMYZWWCGF": "",
                            "L3IBEEOE6UHRRS": "",
                        },
                        "columnSelectFunctions": {
                            "BPZ7OF3CU3BTPJ": {
                                "nodeType": "output",
                                "type": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "start": "float",
                                        "stop": "float",
                                    },
                                },
                                "fromOp": {
                                    "name": "number-bin",
                                    "inputs": {
                                        "in_": {
                                            "nodeType": "output",
                                            "type": "int",
                                            "fromOp": {
                                                "name": "typedDict-pick",
                                                "inputs": {
                                                    "self": {
                                                        "nodeType": "var",
                                                        "type": {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "step": "int",
                                                                "string_col": "string",
                                                                "metric0": "float",
                                                                "metric1": "float",
                                                                "metric2": "float",
                                                                "metric3": "float",
                                                                "metric4": "float",
                                                                "metric5": "float",
                                                                "metric6": "float",
                                                                "metric7": "float",
                                                                "metric8": "float",
                                                                "metric9": "float",
                                                                "metric10": "float",
                                                                "metric11": "float",
                                                                "metric12": "float",
                                                                "metric13": "float",
                                                                "metric14": "float",
                                                                "metric15": "float",
                                                                "metric16": "float",
                                                                "metric17": "float",
                                                                "metric18": "float",
                                                                "metric19": "float",
                                                                "metric20": "float",
                                                                "metric21": "float",
                                                                "metric22": "float",
                                                                "metric23": "float",
                                                                "metric24": "float",
                                                                "metric25": "float",
                                                                "metric26": "float",
                                                                "metric27": "float",
                                                                "metric28": "float",
                                                                "metric29": "float",
                                                                "metric30": "float",
                                                                "metric31": "float",
                                                                "metric32": "float",
                                                                "metric33": "float",
                                                                "metric34": "float",
                                                                "metric35": "float",
                                                                "metric36": "float",
                                                                "metric37": "float",
                                                                "metric38": "float",
                                                                "metric39": "float",
                                                                "metric40": "float",
                                                                "metric41": "float",
                                                                "metric42": "float",
                                                                "metric43": "float",
                                                                "metric44": "float",
                                                                "metric45": "float",
                                                                "metric46": "float",
                                                                "metric47": "float",
                                                                "metric48": "float",
                                                                "metric49": "float",
                                                                "metric50": "float",
                                                                "metric51": "float",
                                                                "metric52": "float",
                                                                "metric53": "float",
                                                                "metric54": "float",
                                                                "metric55": "float",
                                                                "metric56": "float",
                                                                "metric57": "float",
                                                                "metric58": "float",
                                                                "metric59": "float",
                                                                "metric60": "float",
                                                                "metric61": "float",
                                                                "metric62": "float",
                                                                "metric63": "float",
                                                                "metric64": "float",
                                                                "metric65": "float",
                                                                "metric66": "float",
                                                                "metric67": "float",
                                                                "metric68": "float",
                                                                "metric69": "float",
                                                                "metric70": "float",
                                                                "metric71": "float",
                                                                "metric72": "float",
                                                                "metric73": "float",
                                                                "metric74": "float",
                                                                "metric75": "float",
                                                                "metric76": "float",
                                                                "metric77": "float",
                                                                "metric78": "float",
                                                                "metric79": "float",
                                                                "metric80": "float",
                                                                "metric81": "float",
                                                                "metric82": "float",
                                                                "metric83": "float",
                                                                "metric84": "float",
                                                                "metric85": "float",
                                                                "metric86": "float",
                                                                "metric87": "float",
                                                                "metric88": "float",
                                                                "metric89": "float",
                                                                "metric90": "float",
                                                                "metric91": "float",
                                                                "metric92": "float",
                                                                "metric93": "float",
                                                                "metric94": "float",
                                                                "metric95": "float",
                                                                "metric96": "float",
                                                                "metric97": "float",
                                                                "metric98": "float",
                                                                "run_name": "string",
                                                            },
                                                        },
                                                        "varName": "row",
                                                    },
                                                    "key": {
                                                        "nodeType": "const",
                                                        "type": "string",
                                                        "val": "step",
                                                    },
                                                },
                                            },
                                        },
                                        "bin_fn": {
                                            "nodeType": "output",
                                            "type": {
                                                "type": "function",
                                                "inputTypes": {"row": "number"},
                                                "outputType": {
                                                    "type": "typedDict",
                                                    "propertyTypes": {
                                                        "start": "float",
                                                        "stop": "float",
                                                    },
                                                },
                                            },
                                            "fromOp": {
                                                "name": "numbers-binsequal",
                                                "inputs": {
                                                    "arr": {
                                                        "nodeType": "output",
                                                        "type": {
                                                            "type": "list",
                                                            "objectType": "int",
                                                        },
                                                        "fromOp": {
                                                            "name": "get",
                                                            "inputs": {
                                                                "uri": {
                                                                    "nodeType": "const",
                                                                    "type": "string",
                                                                    "val": "local-artifact:///tmp/weave/pytest/weave/tests/test_plot.py::test_actual_config_value (setup)/list/dc9ffba8da33b6d87d72d4525afcd383",
                                                                }
                                                            },
                                                        },
                                                    },
                                                    "bins": {
                                                        "nodeType": "const",
                                                        "type": "int",
                                                        "val": 2,
                                                    },
                                                },
                                            },
                                        },
                                    },
                                },
                            },
                            "3TAMZ9RE8CSQNQ": {
                                "nodeType": "output",
                                "type": "number",
                                "fromOp": {
                                    "name": "numbers-avg",
                                    "inputs": {
                                        "numbers": {
                                            "nodeType": "output",
                                            "type": "float",
                                            "fromOp": {
                                                "name": "typedDict-pick",
                                                "inputs": {
                                                    "self": {
                                                        "nodeType": "var",
                                                        "type": {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "step": "int",
                                                                "string_col": "string",
                                                                "metric0": "float",
                                                                "metric1": "float",
                                                                "metric2": "float",
                                                                "metric3": "float",
                                                                "metric4": "float",
                                                                "metric5": "float",
                                                                "metric6": "float",
                                                                "metric7": "float",
                                                                "metric8": "float",
                                                                "metric9": "float",
                                                                "metric10": "float",
                                                                "metric11": "float",
                                                                "metric12": "float",
                                                                "metric13": "float",
                                                                "metric14": "float",
                                                                "metric15": "float",
                                                                "metric16": "float",
                                                                "metric17": "float",
                                                                "metric18": "float",
                                                                "metric19": "float",
                                                                "metric20": "float",
                                                                "metric21": "float",
                                                                "metric22": "float",
                                                                "metric23": "float",
                                                                "metric24": "float",
                                                                "metric25": "float",
                                                                "metric26": "float",
                                                                "metric27": "float",
                                                                "metric28": "float",
                                                                "metric29": "float",
                                                                "metric30": "float",
                                                                "metric31": "float",
                                                                "metric32": "float",
                                                                "metric33": "float",
                                                                "metric34": "float",
                                                                "metric35": "float",
                                                                "metric36": "float",
                                                                "metric37": "float",
                                                                "metric38": "float",
                                                                "metric39": "float",
                                                                "metric40": "float",
                                                                "metric41": "float",
                                                                "metric42": "float",
                                                                "metric43": "float",
                                                                "metric44": "float",
                                                                "metric45": "float",
                                                                "metric46": "float",
                                                                "metric47": "float",
                                                                "metric48": "float",
                                                                "metric49": "float",
                                                                "metric50": "float",
                                                                "metric51": "float",
                                                                "metric52": "float",
                                                                "metric53": "float",
                                                                "metric54": "float",
                                                                "metric55": "float",
                                                                "metric56": "float",
                                                                "metric57": "float",
                                                                "metric58": "float",
                                                                "metric59": "float",
                                                                "metric60": "float",
                                                                "metric61": "float",
                                                                "metric62": "float",
                                                                "metric63": "float",
                                                                "metric64": "float",
                                                                "metric65": "float",
                                                                "metric66": "float",
                                                                "metric67": "float",
                                                                "metric68": "float",
                                                                "metric69": "float",
                                                                "metric70": "float",
                                                                "metric71": "float",
                                                                "metric72": "float",
                                                                "metric73": "float",
                                                                "metric74": "float",
                                                                "metric75": "float",
                                                                "metric76": "float",
                                                                "metric77": "float",
                                                                "metric78": "float",
                                                                "metric79": "float",
                                                                "metric80": "float",
                                                                "metric81": "float",
                                                                "metric82": "float",
                                                                "metric83": "float",
                                                                "metric84": "float",
                                                                "metric85": "float",
                                                                "metric86": "float",
                                                                "metric87": "float",
                                                                "metric88": "float",
                                                                "metric89": "float",
                                                                "metric90": "float",
                                                                "metric91": "float",
                                                                "metric92": "float",
                                                                "metric93": "float",
                                                                "metric94": "float",
                                                                "metric95": "float",
                                                                "metric96": "float",
                                                                "metric97": "float",
                                                                "metric98": "float",
                                                                "run_name": "string",
                                                            },
                                                        },
                                                        "varName": "row",
                                                    },
                                                    "key": {
                                                        "nodeType": "const",
                                                        "type": "string",
                                                        "val": "metric0",
                                                    },
                                                },
                                            },
                                        }
                                    },
                                },
                            },
                            "V4ZUZP68VEFUE4": {"nodeType": "void", "type": "invalid"},
                            "DX2S5BL68G31XO": {"nodeType": "void", "type": "invalid"},
                            "6YQY0NYGB338WC": {"nodeType": "void", "type": "invalid"},
                            "FSW0FWMRIMFYK6": {"nodeType": "void", "type": "invalid"},
                            "UYPJ7CMYZWWCGF": {"nodeType": "void", "type": "invalid"},
                            "L3IBEEOE6UHRRS": {"nodeType": "void", "type": "invalid"},
                        },
                        "order": [
                            "BPZ7OF3CU3BTPJ",
                            "3TAMZ9RE8CSQNQ",
                            "V4ZUZP68VEFUE4",
                            "DX2S5BL68G31XO",
                            "6YQY0NYGB338WC",
                            "FSW0FWMRIMFYK6",
                            "UYPJ7CMYZWWCGF",
                            "L3IBEEOE6UHRRS",
                        ],
                        "groupBy": [],
                        "sort": [],
                        "pageSize": 10,
                        "page": 0,
                    },
                    "constants": {
                        "mark": "line",
                        "pointShape": None,
                        "lineStyle": None,
                        "label": None,
                    },
                    "uiState": {"pointShape": "expression", "label": "expression"},
                    "dims": {
                        "x": "BPZ7OF3CU3BTPJ",
                        "y": "3TAMZ9RE8CSQNQ",
                        "color": "V4ZUZP68VEFUE4",
                        "label": "DX2S5BL68G31XO",
                        "tooltip": "6YQY0NYGB338WC",
                        "pointSize": "FSW0FWMRIMFYK6",
                        "pointShape": "UYPJ7CMYZWWCGF",
                        "y2": "L3IBEEOE6UHRRS",
                    },
                },
                {
                    "table": {
                        "columns": {
                            "BPZ7OF3CU3BTPJ": {"panelConfig": None, "panelId": ""},
                            "3TAMZ9RE8CSQNQ": {"panelConfig": None, "panelId": ""},
                            "V4ZUZP68VEFUE4": {"panelConfig": None, "panelId": ""},
                            "DX2S5BL68G31XO": {"panelConfig": None, "panelId": ""},
                            "6YQY0NYGB338WC": {"panelConfig": None, "panelId": ""},
                            "FSW0FWMRIMFYK6": {"panelConfig": None, "panelId": ""},
                            "UYPJ7CMYZWWCGF": {"panelConfig": None, "panelId": ""},
                            "L3IBEEOE6UHRRS": {"panelConfig": None, "panelId": ""},
                        },
                        "input_node": {
                            "nodeType": "output",
                            "type": {
                                "type": "ArrowWeaveList",
                                "objectType": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "step": "int",
                                        "string_col": "string",
                                        "metric0": "float",
                                        "metric1": "float",
                                        "metric2": "float",
                                        "metric3": "float",
                                        "metric4": "float",
                                        "metric5": "float",
                                        "metric6": "float",
                                        "metric7": "float",
                                        "metric8": "float",
                                        "metric9": "float",
                                        "metric10": "float",
                                        "metric11": "float",
                                        "metric12": "float",
                                        "metric13": "float",
                                        "metric14": "float",
                                        "metric15": "float",
                                        "metric16": "float",
                                        "metric17": "float",
                                        "metric18": "float",
                                        "metric19": "float",
                                        "metric20": "float",
                                        "metric21": "float",
                                        "metric22": "float",
                                        "metric23": "float",
                                        "metric24": "float",
                                        "metric25": "float",
                                        "metric26": "float",
                                        "metric27": "float",
                                        "metric28": "float",
                                        "metric29": "float",
                                        "metric30": "float",
                                        "metric31": "float",
                                        "metric32": "float",
                                        "metric33": "float",
                                        "metric34": "float",
                                        "metric35": "float",
                                        "metric36": "float",
                                        "metric37": "float",
                                        "metric38": "float",
                                        "metric39": "float",
                                        "metric40": "float",
                                        "metric41": "float",
                                        "metric42": "float",
                                        "metric43": "float",
                                        "metric44": "float",
                                        "metric45": "float",
                                        "metric46": "float",
                                        "metric47": "float",
                                        "metric48": "float",
                                        "metric49": "float",
                                        "metric50": "float",
                                        "metric51": "float",
                                        "metric52": "float",
                                        "metric53": "float",
                                        "metric54": "float",
                                        "metric55": "float",
                                        "metric56": "float",
                                        "metric57": "float",
                                        "metric58": "float",
                                        "metric59": "float",
                                        "metric60": "float",
                                        "metric61": "float",
                                        "metric62": "float",
                                        "metric63": "float",
                                        "metric64": "float",
                                        "metric65": "float",
                                        "metric66": "float",
                                        "metric67": "float",
                                        "metric68": "float",
                                        "metric69": "float",
                                        "metric70": "float",
                                        "metric71": "float",
                                        "metric72": "float",
                                        "metric73": "float",
                                        "metric74": "float",
                                        "metric75": "float",
                                        "metric76": "float",
                                        "metric77": "float",
                                        "metric78": "float",
                                        "metric79": "float",
                                        "metric80": "float",
                                        "metric81": "float",
                                        "metric82": "float",
                                        "metric83": "float",
                                        "metric84": "float",
                                        "metric85": "float",
                                        "metric86": "float",
                                        "metric87": "float",
                                        "metric88": "float",
                                        "metric89": "float",
                                        "metric90": "float",
                                        "metric91": "float",
                                        "metric92": "float",
                                        "metric93": "float",
                                        "metric94": "float",
                                        "metric95": "float",
                                        "metric96": "float",
                                        "metric97": "float",
                                        "metric98": "float",
                                        "run_name": "string",
                                    },
                                },
                            },
                            "fromOp": {
                                "name": "RunSegment-experiment",
                                "inputs": {
                                    "self": {
                                        "nodeType": "output",
                                        "type": {
                                            "type": "RunSegment",
                                            "_is_object": True,
                                            "prior_run_ref": "string",
                                            "prior_run_branch_index": "int",
                                            "metrics": {
                                                "type": "ArrowWeaveList",
                                                "objectType": {
                                                    "type": "typedDict",
                                                    "propertyTypes": {
                                                        "step": "int",
                                                        "string_col": "string",
                                                        "metric0": "float",
                                                        "metric1": "float",
                                                        "metric2": "float",
                                                        "metric3": "float",
                                                        "metric4": "float",
                                                        "metric5": "float",
                                                        "metric6": "float",
                                                        "metric7": "float",
                                                        "metric8": "float",
                                                        "metric9": "float",
                                                        "metric10": "float",
                                                        "metric11": "float",
                                                        "metric12": "float",
                                                        "metric13": "float",
                                                        "metric14": "float",
                                                        "metric15": "float",
                                                        "metric16": "float",
                                                        "metric17": "float",
                                                        "metric18": "float",
                                                        "metric19": "float",
                                                        "metric20": "float",
                                                        "metric21": "float",
                                                        "metric22": "float",
                                                        "metric23": "float",
                                                        "metric24": "float",
                                                        "metric25": "float",
                                                        "metric26": "float",
                                                        "metric27": "float",
                                                        "metric28": "float",
                                                        "metric29": "float",
                                                        "metric30": "float",
                                                        "metric31": "float",
                                                        "metric32": "float",
                                                        "metric33": "float",
                                                        "metric34": "float",
                                                        "metric35": "float",
                                                        "metric36": "float",
                                                        "metric37": "float",
                                                        "metric38": "float",
                                                        "metric39": "float",
                                                        "metric40": "float",
                                                        "metric41": "float",
                                                        "metric42": "float",
                                                        "metric43": "float",
                                                        "metric44": "float",
                                                        "metric45": "float",
                                                        "metric46": "float",
                                                        "metric47": "float",
                                                        "metric48": "float",
                                                        "metric49": "float",
                                                        "metric50": "float",
                                                        "metric51": "float",
                                                        "metric52": "float",
                                                        "metric53": "float",
                                                        "metric54": "float",
                                                        "metric55": "float",
                                                        "metric56": "float",
                                                        "metric57": "float",
                                                        "metric58": "float",
                                                        "metric59": "float",
                                                        "metric60": "float",
                                                        "metric61": "float",
                                                        "metric62": "float",
                                                        "metric63": "float",
                                                        "metric64": "float",
                                                        "metric65": "float",
                                                        "metric66": "float",
                                                        "metric67": "float",
                                                        "metric68": "float",
                                                        "metric69": "float",
                                                        "metric70": "float",
                                                        "metric71": "float",
                                                        "metric72": "float",
                                                        "metric73": "float",
                                                        "metric74": "float",
                                                        "metric75": "float",
                                                        "metric76": "float",
                                                        "metric77": "float",
                                                        "metric78": "float",
                                                        "metric79": "float",
                                                        "metric80": "float",
                                                        "metric81": "float",
                                                        "metric82": "float",
                                                        "metric83": "float",
                                                        "metric84": "float",
                                                        "metric85": "float",
                                                        "metric86": "float",
                                                        "metric87": "float",
                                                        "metric88": "float",
                                                        "metric89": "float",
                                                        "metric90": "float",
                                                        "metric91": "float",
                                                        "metric92": "float",
                                                        "metric93": "float",
                                                        "metric94": "float",
                                                        "metric95": "float",
                                                        "metric96": "float",
                                                        "metric97": "float",
                                                        "metric98": "float",
                                                    },
                                                },
                                            },
                                            "run_name": "string",
                                        },
                                        "fromOp": {
                                            "name": "get",
                                            "inputs": {
                                                "uri": {
                                                    "nodeType": "const",
                                                    "type": "string",
                                                    "val": "local-artifact:///tmp/weave/pytest/weave/tests/test_plot.py::test_actual_config_value (setup)/RunSegment/164b91e6b2832f79254c6da17e094563",
                                                }
                                            },
                                        },
                                    }
                                },
                            },
                        },
                        "autoColumns": False,
                        "preFilterFunction": {"nodeType": "void", "type": "invalid"},
                        "columnNames": {
                            "BPZ7OF3CU3BTPJ": "",
                            "3TAMZ9RE8CSQNQ": "",
                            "V4ZUZP68VEFUE4": "",
                            "DX2S5BL68G31XO": "",
                            "6YQY0NYGB338WC": "",
                            "FSW0FWMRIMFYK6": "",
                            "UYPJ7CMYZWWCGF": "",
                            "L3IBEEOE6UHRRS": "",
                        },
                        "columnSelectFunctions": {
                            "BPZ7OF3CU3BTPJ": {
                                "nodeType": "output",
                                "type": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "start": "float",
                                        "stop": "float",
                                    },
                                },
                                "fromOp": {
                                    "name": "number-bin",
                                    "inputs": {
                                        "in_": {
                                            "nodeType": "output",
                                            "type": "int",
                                            "fromOp": {
                                                "name": "typedDict-pick",
                                                "inputs": {
                                                    "self": {
                                                        "nodeType": "var",
                                                        "type": {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "step": "int",
                                                                "string_col": "string",
                                                                "metric0": "float",
                                                                "metric1": "float",
                                                                "metric2": "float",
                                                                "metric3": "float",
                                                                "metric4": "float",
                                                                "metric5": "float",
                                                                "metric6": "float",
                                                                "metric7": "float",
                                                                "metric8": "float",
                                                                "metric9": "float",
                                                                "metric10": "float",
                                                                "metric11": "float",
                                                                "metric12": "float",
                                                                "metric13": "float",
                                                                "metric14": "float",
                                                                "metric15": "float",
                                                                "metric16": "float",
                                                                "metric17": "float",
                                                                "metric18": "float",
                                                                "metric19": "float",
                                                                "metric20": "float",
                                                                "metric21": "float",
                                                                "metric22": "float",
                                                                "metric23": "float",
                                                                "metric24": "float",
                                                                "metric25": "float",
                                                                "metric26": "float",
                                                                "metric27": "float",
                                                                "metric28": "float",
                                                                "metric29": "float",
                                                                "metric30": "float",
                                                                "metric31": "float",
                                                                "metric32": "float",
                                                                "metric33": "float",
                                                                "metric34": "float",
                                                                "metric35": "float",
                                                                "metric36": "float",
                                                                "metric37": "float",
                                                                "metric38": "float",
                                                                "metric39": "float",
                                                                "metric40": "float",
                                                                "metric41": "float",
                                                                "metric42": "float",
                                                                "metric43": "float",
                                                                "metric44": "float",
                                                                "metric45": "float",
                                                                "metric46": "float",
                                                                "metric47": "float",
                                                                "metric48": "float",
                                                                "metric49": "float",
                                                                "metric50": "float",
                                                                "metric51": "float",
                                                                "metric52": "float",
                                                                "metric53": "float",
                                                                "metric54": "float",
                                                                "metric55": "float",
                                                                "metric56": "float",
                                                                "metric57": "float",
                                                                "metric58": "float",
                                                                "metric59": "float",
                                                                "metric60": "float",
                                                                "metric61": "float",
                                                                "metric62": "float",
                                                                "metric63": "float",
                                                                "metric64": "float",
                                                                "metric65": "float",
                                                                "metric66": "float",
                                                                "metric67": "float",
                                                                "metric68": "float",
                                                                "metric69": "float",
                                                                "metric70": "float",
                                                                "metric71": "float",
                                                                "metric72": "float",
                                                                "metric73": "float",
                                                                "metric74": "float",
                                                                "metric75": "float",
                                                                "metric76": "float",
                                                                "metric77": "float",
                                                                "metric78": "float",
                                                                "metric79": "float",
                                                                "metric80": "float",
                                                                "metric81": "float",
                                                                "metric82": "float",
                                                                "metric83": "float",
                                                                "metric84": "float",
                                                                "metric85": "float",
                                                                "metric86": "float",
                                                                "metric87": "float",
                                                                "metric88": "float",
                                                                "metric89": "float",
                                                                "metric90": "float",
                                                                "metric91": "float",
                                                                "metric92": "float",
                                                                "metric93": "float",
                                                                "metric94": "float",
                                                                "metric95": "float",
                                                                "metric96": "float",
                                                                "metric97": "float",
                                                                "metric98": "float",
                                                                "run_name": "string",
                                                            },
                                                        },
                                                        "varName": "row",
                                                    },
                                                    "key": {
                                                        "nodeType": "const",
                                                        "type": "string",
                                                        "val": "step",
                                                    },
                                                },
                                            },
                                        },
                                        "bin_fn": {
                                            "nodeType": "output",
                                            "type": {
                                                "type": "function",
                                                "inputTypes": {"row": "number"},
                                                "outputType": {
                                                    "type": "typedDict",
                                                    "propertyTypes": {
                                                        "start": "float",
                                                        "stop": "float",
                                                    },
                                                },
                                            },
                                            "fromOp": {
                                                "name": "numbers-binsequal",
                                                "inputs": {
                                                    "arr": {
                                                        "nodeType": "output",
                                                        "type": {
                                                            "type": "list",
                                                            "objectType": "int",
                                                        },
                                                        "fromOp": {
                                                            "name": "get",
                                                            "inputs": {
                                                                "uri": {
                                                                    "nodeType": "const",
                                                                    "type": "string",
                                                                    "val": "local-artifact:///tmp/weave/pytest/weave/tests/test_plot.py::test_actual_config_value (setup)/list/dc9ffba8da33b6d87d72d4525afcd383",
                                                                }
                                                            },
                                                        },
                                                    },
                                                    "bins": {
                                                        "nodeType": "const",
                                                        "type": "int",
                                                        "val": 2,
                                                    },
                                                },
                                            },
                                        },
                                    },
                                },
                            },
                            "3TAMZ9RE8CSQNQ": {
                                "nodeType": "output",
                                "type": "number",
                                "fromOp": {
                                    "name": "numbers-min",
                                    "inputs": {
                                        "numbers": {
                                            "nodeType": "output",
                                            "type": "float",
                                            "fromOp": {
                                                "name": "typedDict-pick",
                                                "inputs": {
                                                    "self": {
                                                        "nodeType": "var",
                                                        "type": {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "step": "int",
                                                                "string_col": "string",
                                                                "metric0": "float",
                                                                "metric1": "float",
                                                                "metric2": "float",
                                                                "metric3": "float",
                                                                "metric4": "float",
                                                                "metric5": "float",
                                                                "metric6": "float",
                                                                "metric7": "float",
                                                                "metric8": "float",
                                                                "metric9": "float",
                                                                "metric10": "float",
                                                                "metric11": "float",
                                                                "metric12": "float",
                                                                "metric13": "float",
                                                                "metric14": "float",
                                                                "metric15": "float",
                                                                "metric16": "float",
                                                                "metric17": "float",
                                                                "metric18": "float",
                                                                "metric19": "float",
                                                                "metric20": "float",
                                                                "metric21": "float",
                                                                "metric22": "float",
                                                                "metric23": "float",
                                                                "metric24": "float",
                                                                "metric25": "float",
                                                                "metric26": "float",
                                                                "metric27": "float",
                                                                "metric28": "float",
                                                                "metric29": "float",
                                                                "metric30": "float",
                                                                "metric31": "float",
                                                                "metric32": "float",
                                                                "metric33": "float",
                                                                "metric34": "float",
                                                                "metric35": "float",
                                                                "metric36": "float",
                                                                "metric37": "float",
                                                                "metric38": "float",
                                                                "metric39": "float",
                                                                "metric40": "float",
                                                                "metric41": "float",
                                                                "metric42": "float",
                                                                "metric43": "float",
                                                                "metric44": "float",
                                                                "metric45": "float",
                                                                "metric46": "float",
                                                                "metric47": "float",
                                                                "metric48": "float",
                                                                "metric49": "float",
                                                                "metric50": "float",
                                                                "metric51": "float",
                                                                "metric52": "float",
                                                                "metric53": "float",
                                                                "metric54": "float",
                                                                "metric55": "float",
                                                                "metric56": "float",
                                                                "metric57": "float",
                                                                "metric58": "float",
                                                                "metric59": "float",
                                                                "metric60": "float",
                                                                "metric61": "float",
                                                                "metric62": "float",
                                                                "metric63": "float",
                                                                "metric64": "float",
                                                                "metric65": "float",
                                                                "metric66": "float",
                                                                "metric67": "float",
                                                                "metric68": "float",
                                                                "metric69": "float",
                                                                "metric70": "float",
                                                                "metric71": "float",
                                                                "metric72": "float",
                                                                "metric73": "float",
                                                                "metric74": "float",
                                                                "metric75": "float",
                                                                "metric76": "float",
                                                                "metric77": "float",
                                                                "metric78": "float",
                                                                "metric79": "float",
                                                                "metric80": "float",
                                                                "metric81": "float",
                                                                "metric82": "float",
                                                                "metric83": "float",
                                                                "metric84": "float",
                                                                "metric85": "float",
                                                                "metric86": "float",
                                                                "metric87": "float",
                                                                "metric88": "float",
                                                                "metric89": "float",
                                                                "metric90": "float",
                                                                "metric91": "float",
                                                                "metric92": "float",
                                                                "metric93": "float",
                                                                "metric94": "float",
                                                                "metric95": "float",
                                                                "metric96": "float",
                                                                "metric97": "float",
                                                                "metric98": "float",
                                                                "run_name": "string",
                                                            },
                                                        },
                                                        "varName": "row",
                                                    },
                                                    "key": {
                                                        "nodeType": "const",
                                                        "type": "string",
                                                        "val": "metric0",
                                                    },
                                                },
                                            },
                                        }
                                    },
                                },
                            },
                            "V4ZUZP68VEFUE4": {"nodeType": "void", "type": "invalid"},
                            "DX2S5BL68G31XO": {"nodeType": "void", "type": "invalid"},
                            "6YQY0NYGB338WC": {"nodeType": "void", "type": "invalid"},
                            "FSW0FWMRIMFYK6": {"nodeType": "void", "type": "invalid"},
                            "UYPJ7CMYZWWCGF": {"nodeType": "void", "type": "invalid"},
                            "L3IBEEOE6UHRRS": {
                                "nodeType": "output",
                                "type": "number",
                                "fromOp": {
                                    "name": "numbers-max",
                                    "inputs": {
                                        "numbers": {
                                            "nodeType": "output",
                                            "type": "float",
                                            "fromOp": {
                                                "name": "typedDict-pick",
                                                "inputs": {
                                                    "self": {
                                                        "nodeType": "var",
                                                        "type": {
                                                            "type": "typedDict",
                                                            "propertyTypes": {
                                                                "step": "int",
                                                                "string_col": "string",
                                                                "metric0": "float",
                                                                "metric1": "float",
                                                                "metric2": "float",
                                                                "metric3": "float",
                                                                "metric4": "float",
                                                                "metric5": "float",
                                                                "metric6": "float",
                                                                "metric7": "float",
                                                                "metric8": "float",
                                                                "metric9": "float",
                                                                "metric10": "float",
                                                                "metric11": "float",
                                                                "metric12": "float",
                                                                "metric13": "float",
                                                                "metric14": "float",
                                                                "metric15": "float",
                                                                "metric16": "float",
                                                                "metric17": "float",
                                                                "metric18": "float",
                                                                "metric19": "float",
                                                                "metric20": "float",
                                                                "metric21": "float",
                                                                "metric22": "float",
                                                                "metric23": "float",
                                                                "metric24": "float",
                                                                "metric25": "float",
                                                                "metric26": "float",
                                                                "metric27": "float",
                                                                "metric28": "float",
                                                                "metric29": "float",
                                                                "metric30": "float",
                                                                "metric31": "float",
                                                                "metric32": "float",
                                                                "metric33": "float",
                                                                "metric34": "float",
                                                                "metric35": "float",
                                                                "metric36": "float",
                                                                "metric37": "float",
                                                                "metric38": "float",
                                                                "metric39": "float",
                                                                "metric40": "float",
                                                                "metric41": "float",
                                                                "metric42": "float",
                                                                "metric43": "float",
                                                                "metric44": "float",
                                                                "metric45": "float",
                                                                "metric46": "float",
                                                                "metric47": "float",
                                                                "metric48": "float",
                                                                "metric49": "float",
                                                                "metric50": "float",
                                                                "metric51": "float",
                                                                "metric52": "float",
                                                                "metric53": "float",
                                                                "metric54": "float",
                                                                "metric55": "float",
                                                                "metric56": "float",
                                                                "metric57": "float",
                                                                "metric58": "float",
                                                                "metric59": "float",
                                                                "metric60": "float",
                                                                "metric61": "float",
                                                                "metric62": "float",
                                                                "metric63": "float",
                                                                "metric64": "float",
                                                                "metric65": "float",
                                                                "metric66": "float",
                                                                "metric67": "float",
                                                                "metric68": "float",
                                                                "metric69": "float",
                                                                "metric70": "float",
                                                                "metric71": "float",
                                                                "metric72": "float",
                                                                "metric73": "float",
                                                                "metric74": "float",
                                                                "metric75": "float",
                                                                "metric76": "float",
                                                                "metric77": "float",
                                                                "metric78": "float",
                                                                "metric79": "float",
                                                                "metric80": "float",
                                                                "metric81": "float",
                                                                "metric82": "float",
                                                                "metric83": "float",
                                                                "metric84": "float",
                                                                "metric85": "float",
                                                                "metric86": "float",
                                                                "metric87": "float",
                                                                "metric88": "float",
                                                                "metric89": "float",
                                                                "metric90": "float",
                                                                "metric91": "float",
                                                                "metric92": "float",
                                                                "metric93": "float",
                                                                "metric94": "float",
                                                                "metric95": "float",
                                                                "metric96": "float",
                                                                "metric97": "float",
                                                                "metric98": "float",
                                                                "run_name": "string",
                                                            },
                                                        },
                                                        "varName": "row",
                                                    },
                                                    "key": {
                                                        "nodeType": "const",
                                                        "type": "string",
                                                        "val": "metric0",
                                                    },
                                                },
                                            },
                                        }
                                    },
                                },
                            },
                        },
                        "order": [
                            "BPZ7OF3CU3BTPJ",
                            "3TAMZ9RE8CSQNQ",
                            "V4ZUZP68VEFUE4",
                            "DX2S5BL68G31XO",
                            "6YQY0NYGB338WC",
                            "FSW0FWMRIMFYK6",
                            "UYPJ7CMYZWWCGF",
                            "L3IBEEOE6UHRRS",
                        ],
                        "groupBy": [],
                        "sort": [],
                        "pageSize": 10,
                        "page": 0,
                    },
                    "constants": {
                        "mark": "area",
                        "pointShape": None,
                        "lineStyle": None,
                        "label": None,
                    },
                    "uiState": {"pointShape": "expression", "label": "expression"},
                    "dims": {
                        "x": "BPZ7OF3CU3BTPJ",
                        "y": "3TAMZ9RE8CSQNQ",
                        "color": "V4ZUZP68VEFUE4",
                        "label": "DX2S5BL68G31XO",
                        "tooltip": "6YQY0NYGB338WC",
                        "pointSize": "FSW0FWMRIMFYK6",
                        "pointShape": "UYPJ7CMYZWWCGF",
                        "y2": "L3IBEEOE6UHRRS",
                    },
                },
            ],
            "axisSettings": {
                "color": {"noLabels": False, "noTitle": False, "noTicks": False},
                "x": {"noLabels": False, "noTitle": False, "noTicks": False},
                "y": {"noLabels": False, "noTitle": False, "noTicks": False},
            },
            "legendSettings": {
                "x": {"noLegend": False},
                "y": {"noLegend": False},
                "color": {"noLegend": False},
            },
            "configOptionsExpanded": {
                "x": False,
                "y": False,
                "label": False,
                "tooltip": False,
                "mark": False,
            },
            "configVersion": 7,
        },
    }


def test_panel_plot_scale_serialization():
    # checks a problem case where scale would not be serialized correctly as an AxisScale object
    plot = weave.legacy.weave.panels.Plot(
        [1, 2, 3, 4],
        x=lambda row: row,
        x_title="x",
        y=lambda row: row**2,
        y_title="y",
        y_axis_type="log",
    )
    type = types.TypeRegistry.type_of(plot)
    serialized = type.to_dict()
    deserialized = types.TypeRegistry.type_from_dict(serialized)
    assert deserialized.assign_type(type) and type.assign_type(deserialized)



import datetime
import typing

import pyarrow as pa
import pytest
from pyarrow import compute as pc

from weave.legacy.weave import api as weave
from weave.legacy.weave import box, dispatch, errors, ops, weave_internal
from weave.legacy.weave import ops_arrow as arrow
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging import (
    make_tag_getter_op,
    tag_store,
    tagged_value_type,
)
from weave.legacy.weave.ops_arrow import arraylist_ops, convert_ops, util
from weave.legacy.weave.ops_domain import run_ops
from weave.legacy.weave.ops_domain import wb_domain_types as wdt
from weave.legacy.weave.ops_primitives import Boolean, Number, date, dict_, list_

string_ops_test_cases = [
    ("eq-scalar", lambda x: x == "bc", [True, False, False]),
    ("ne-scalar", lambda x: x != "bc", [False, True, True]),
    ("contains-scalar", lambda x: x.__contains__("b"), [True, False, False]),
    ("in-scalar", lambda x: x.in_("bcd"), [True, True, False]),
    ("len-scalar", lambda x: x.len(), [2, 2, 2]),
    ("add-scalar", lambda x: x + "q", ["bcq", "cdq", "dfq"]),
    ("append-scalar", lambda x: x.append("qq"), ["bcqq", "cdqq", "dfqq"]),
    ("prepend-scalar", lambda x: x.prepend("qq"), ["qqbc", "qqcd", "qqdf"]),
    ("split-scalar", lambda x: x.split("c"), [["b", ""], ["", "d"], ["df"]]),
    (
        "partition-scalar",
        lambda x: x.partition("c"),
        [["b", "c", ""], ["", "c", "d"], ["df", "", ""]],
    ),
    ("startswith-scalar", lambda x: x.startswith("b"), [True, False, False]),
    ("endswith-scalar", lambda x: x.endswith("f"), [False, False, True]),
    ("replace-scalar", lambda x: x.replace("c", "q"), ["bq", "qd", "df"]),
    ("nest-list", lambda x: list_.make_list(a=x), [["bc"], ["cd"], ["df"]]),
    ("nest-dict", lambda x: dict_(a=x), [{"a": "bc"}, {"a": "cd"}, {"a": "df"}]),
]


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    string_ops_test_cases
    + [
        ("pick", lambda x: dict_(bc=1, cd=2, df=3)[x], [1, 2, 3]),
    ],
)
def test_arrow_vectorizer_string_scalar(name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow(["bc", "cd", "df"]))
    fn = weave_internal.define_fn({"x": weave.types.String()}, weave_func).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert list(weave.use(called)) == expected_output


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    string_ops_test_cases,
)
def test_arrow_vectorizer_string_scalar_tagged(name, weave_func, expected_output):
    expected_value_type = weave.type_of(expected_output[0])

    list = ["bc", "cd", "df"]
    for i, elem in enumerate(list):
        taggable = box.box(elem)
        list[i] = tag_store.add_tags(taggable, {"mytag": f"test{i + 1}"})

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("mytag", types.String())

    awl = arrow.to_arrow(list)
    l = weave.save(awl)
    fn = weave_internal.define_fn({"x": awl.object_type}, weave_func).val
    vec_fn = arrow.vectorize(fn)

    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_notags() == expected_output

    item_node = arrow.ops.index(called, 0)

    def make_exp_tag(t: types.Type):
        return tagged_value_type.TaggedValueType(
            types.TypedDict({"mytag": types.String()}),
            t,
        )

    expected_type_obj_type = make_exp_tag(expected_value_type)
    # The general test here is not general enough to check these properly.
    # Specifcially, the general test assumes the vecotrize lambdas contians all
    # tag flow ops. However, list and dict constructors are explicitly (and
    # correctly) not tag flow ops and therefore the tags are only present on the
    # elements of the list and dict.
    if name == "nest-dict":
        item_node = item_node["a"]
        expected_type_obj_type = types.TypedDict(
            {
                "a": make_exp_tag(expected_value_type.property_types["a"]),
                **expected_value_type.property_types,
            }
        )
    elif name == "nest-list":
        item_node = item_node[0]
        expected_type_obj_type = types.List(
            make_exp_tag(expected_value_type.object_type)
        )

    # check that tags are propagated
    assert weave.use(tag_getter_op(item_node)) == "test1"

    # NOTE: This optionality is needed because some arrow ops eagerly declare
    # optional returns. See number.py and string.py for commentary on the subject.
    assert arrow.ArrowWeaveListType(types.optional(expected_type_obj_type)).assign_type(
        called.type
    )


string_alnum_test_cases = [
    (
        "isAlpha-scalar",
        lambda x: x.isalpha(),
        [False, True, False, False, False, False],
    ),
    (
        "isNumeric-scalar",
        lambda x: x.isnumeric(),
        [False, False, False, True, False, False],
    ),
    (
        "isAlnum-scalar",
        lambda x: x.isalnum(),
        [False, True, True, True, False, False],
    ),
    (
        "lower-scalar",
        lambda x: x.lower(),
        ["b22?c", "cd", "df2", "212", "", "?>!@#"],
    ),
    (
        "upper-scalar",
        lambda x: x.upper(),
        ["B22?C", "CD", "DF2", "212", "", "?>!@#"],
    ),
    (
        "slice-scalar",
        lambda x: x.slice(1, 2),
        ["2", "d", "F", "1", "", ">"],
    ),
]


@pytest.mark.parametrize("name,weave_func,expected_output", string_alnum_test_cases)
def test_arrow_vectorizer_string_alnum(name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow(["B22?c", "cd", "DF2", "212", "", "?>!@#"]))
    fn = weave_internal.define_fn({"x": weave.types.String()}, weave_func).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_raw() == expected_output


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    string_alnum_test_cases,
)
def test_arrow_vectorizer_string_alnum_tagged(name, weave_func, expected_output):
    # NOTE: This optionality is needed because some arrow ops eagerly declare
    # optional returns. See number.py and string.py for commentary on the subject.
    expected_value_type = types.optional(weave.type_of(expected_output[0]))

    list = ["B22?c", "cd", "DF2", "212", "", "?>!@#"]
    for i, elem in enumerate(list):
        taggable = box.box(elem)
        list[i] = tag_store.add_tags(taggable, {"mytag": f"test{i + 1}"})

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("mytag", types.String())

    awl = arrow.to_arrow(list)
    l = weave.save(awl)
    fn = weave_internal.define_fn({"x": awl.object_type}, weave_func).val
    vec_fn = arrow.vectorize(fn)

    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_notags() == expected_output

    # check that tags are propagated
    assert weave.use(tag_getter_op(arrow.ops.index(called, 0))) == "test1"

    assert arrow.ArrowWeaveListType(
        tagged_value_type.TaggedValueType(
            types.TypedDict({"mytag": types.String()}),
            expected_value_type,
        )
    ).assign_type(called.type)


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    [
        ("strip-scalar", lambda x: x.strip(), ["c", "cd", "DF2", "212", ""]),
        ("lstrip-scalar", lambda x: x.lstrip(), ["c ", "cd", "DF2", "212 ", ""]),
        ("rstrip-scalar", lambda x: x.rstrip(), ["  c", "cd", " DF2", "212", ""]),
    ],
)
def test_arrow_vectorizer_string_strip(name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow(["  c ", "cd", " DF2", "212 ", ""]))
    fn = weave_internal.define_fn({"x": weave.types.String()}, weave_func).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_raw() == expected_output


number_ops_test_cases = [
    ("add", lambda x: x + 2, [3.0, 4.0, 5.0]),
    ("add-vec", lambda x: x + x, [2.0, 4.0, 6.0]),
    ("subtract", lambda x: x - 1, [0.0, 1.0, 2.0]),
    ("multiply", lambda x: x * 2, [2.0, 4.0, 6.0]),
    ("divide", lambda x: x / 2, [0.5, 1.0, 1.5]),
    ("pow", lambda x: x**2, [1.0, 4.0, 9.0]),
    ("ne", lambda x: x != 2, [True, False, True]),
    ("eq", lambda x: x == 2, [False, True, False]),
    ("gt", lambda x: x > 2, [False, False, True]),
    ("lt", lambda x: x < 2, [True, False, False]),
    ("ge", lambda x: x >= 2, [False, True, True]),
    ("le", lambda x: x <= 2, [True, True, False]),
    ("neg", lambda x: -x, [-1.0, -2.0, -3.0]),
    ("toString", lambda x: x.toString(), ["1", "2", "3"]),
]


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    number_ops_test_cases,
)
def test_arrow_vectorizer_number_ops(name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow([1.0, 2.0, 3.0]))

    fn = weave_internal.define_fn({"x": weave.types.Float()}, weave_func).val

    vec_fn = arrow.vectorize(fn)

    # TODO:  make it nicer to call vec_fn, we shouldn't need to jump through hoops here

    called = weave_internal.call_fn(vec_fn, {"x": l})

    assert weave.use(called).to_pylist_raw() == expected_output


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    number_ops_test_cases,
)
def test_arrow_vectorizer_number_ops_tagged(name, weave_func, expected_output):
    expected_value_type = weave.type_of(expected_output[0])

    # This special condition is needed because the expected output is a list of
    # booleans is optional booleans. See the comment at the top of `ops_arrow/number.py`
    # for more details.
    if types.Boolean().assign_type(expected_value_type):
        expected_value_type = types.optional(expected_value_type)

    list = [1.0, 2.0, 3.0]
    for i, elem in enumerate(list):
        taggable = box.box(elem)
        list[i] = tag_store.add_tags(taggable, {"mytag": f"test{i + 1}"})

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("mytag", types.String())

    awl = arrow.to_arrow(list)
    l = weave.save(awl)
    fn = weave_internal.define_fn({"x": awl.object_type}, weave_func).val
    vec_fn = arrow.vectorize(fn)

    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_notags() == expected_output

    # check that tags are propagated
    assert weave.use(tag_getter_op(arrow.ops.index(called, 0))) == "test1"

    assert arrow.ArrowWeaveListType(
        tagged_value_type.TaggedValueType(
            types.TypedDict({"mytag": types.String()}),
            expected_value_type,
        )
    ).assign_type(called.type)


string_vector_ops_test_cases = [
    ("eq-vector", lambda x, y: x == y, [False, False, True]),
    ("ne-vector", lambda x, y: x != y, [True, True, False]),
    ("contains-vector", lambda x, y: x.__contains__(y), [False, False, True]),
    ("in-vector", lambda x, y: x.in_(y), [True, False, True]),
    ("add-vector", lambda x, y: x + y, ["bccbc", "cdaef", "dfdf"]),
    ("append-vector", lambda x, y: x.append(y), ["bccbc", "cdaef", "dfdf"]),
    ("prepend-vector", lambda x, y: x.prepend(y), ["cbcbc", "aefcd", "dfdf"]),
    ("split-vector", lambda x, y: y.split(x), [["c", ""], ["aef"], ["", ""]]),
    (
        "partition-vector",
        lambda x, y: y.partition(x),
        [["c", "bc", ""], ["aef", "", ""], ["", "df", ""]],
    ),
    (
        "startswith-vector",
        lambda x, y: y.startswith(x),
        [False, False, True],
    ),
    ("endswith-vector", lambda x, y: y.endswith(x), [True, False, True]),
]


@pytest.mark.parametrize(
    "name,weave_func,expected_output", string_vector_ops_test_cases
)
def test_arrow_vectorizer_string_vector(name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow(["bc", "cd", "df"]))
    l2 = weave.save(arrow.to_arrow(["cbc", "aef", "df"]))

    fn = weave_internal.define_fn(
        {"x": weave.types.String(), "y": weave.types.String()}, weave_func
    ).val

    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l, "y": l2})
    assert weave.use(called).to_pylist_raw() == expected_output


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    [("string-toNumber", lambda x: x.toNumber(), [123, 234, 456, None])],
)
def test_arrow_vectorizer_string_to_int(name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow(["123", "234", "456", "a"]))

    fn = weave_internal.define_fn({"x": weave.types.String()}, weave_func).val

    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    weave.use(called).to_pylist_raw() == expected_output


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    string_vector_ops_test_cases,
)
def test_arrow_vectorizer_string_vector_ops_tagged(name, weave_func, expected_output):
    # NOTE: This optionality is needed because some arrow ops eagerly declare
    # optional returns. See number.py and string.py for commentary on the subject.
    expected_value_type = types.optional(weave.type_of(expected_output[0]))

    list = ["bc", "cd", "df"]
    for i, elem in enumerate(list):
        taggable = box.box(elem)
        list[i] = tag_store.add_tags(taggable, {"mytag": f"test{i + 1}"})

    list2 = ["cbc", "aef", "df"]
    for i, elem in enumerate(list2):
        taggable = box.box(elem)
        list2[i] = tag_store.add_tags(taggable, {"mytag2": f"test{i + 1}"})

    awl = arrow.to_arrow(list)
    awl2 = arrow.to_arrow(list2)

    l = weave.save(awl)
    l2 = weave.save(awl2)
    fn = weave_internal.define_fn(
        {"x": awl.object_type, "y": awl2.object_type}, weave_func
    ).val
    vec_fn = arrow.vectorize(fn)

    called = weave_internal.call_fn(vec_fn, {"x": l, "y": l2})
    assert weave.use(called).to_pylist_notags() == expected_output

    # if x is first, expect x tag to be propagated. if y is first, expect y tag to be propagated.
    tag_name: typing.Optional[str] = None
    for name, node in called.iteritems_op_inputs():
        if name == "self":
            tag_name = next(k for k in node.type.object_type.tag.property_types)

    if tag_name is None:
        raise ValueError("tag_getter_op not found")

    tag_getter_op = make_tag_getter_op.make_tag_getter_op(
        tag_name,
        types.String(),
    )

    # check that tags are propagated
    assert weave.use(tag_getter_op(arrow.ops.index(called, 0))) == "test1"

    assert arrow.ArrowWeaveListType(
        tagged_value_type.TaggedValueType(
            types.TypedDict({tag_name: types.String()}),
            expected_value_type,
        )
    ).assign_type(called.type)


nullable_number_ops_test_cases = [
    ("add", lambda x: x + 2, [3.0, None, 5.0]),
    ("add-vec", lambda x: x + x, [2.0, None, 6.0]),
    ("subtract", lambda x: x - 1, [0.0, None, 2.0]),
    ("multiply", lambda x: x * 2, [2.0, None, 6.0]),
    ("divide", lambda x: x / 2, [0.5, None, 1.5]),
    ("pow", lambda x: x**2, [1.0, None, 9.0]),
    ("ne", lambda x: x != 2, [True, True, True]),
    ("eq", lambda x: x == 2, [False, False, False]),
    ("gt", lambda x: x > 2, [False, None, True]),
    ("lt", lambda x: x < 2, [True, None, False]),
    ("ge", lambda x: x >= 2, [False, None, True]),
    ("le", lambda x: x <= 2, [True, None, False]),
    ("neg", lambda x: -x, [-1.0, None, -3.0]),
    ("toString", lambda x: x.toString(), ["1", None, "3"]),
]


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    nullable_number_ops_test_cases,
)
def test_arrow_vectorizer_nullable_number_ops(name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow([1.0, None, 3.0]))

    fn = weave_internal.define_fn({"x": weave.types.Float()}, weave_func).val

    vec_fn = arrow.vectorize(fn)

    # TODO:  make it nicer to call vec_fn, we shouldn't need to jump through hoops here

    called = weave_internal.call_fn(vec_fn, {"x": l})

    assert weave.use(called).to_pylist_raw() == expected_output


string_ops_nullable_test_cases = [
    ("eq-scalar", lambda x: x == "bc", [True, False, False]),
    ("ne-scalar", lambda x: x != "bc", [False, True, True]),
    ("contains-scalar", lambda x: x.__contains__("b"), [True, None, False]),
    ("in-scalar", lambda x: x.in_("bcd"), [True, None, False]),
    ("len-scalar", lambda x: x.len(), [2, None, 2]),
    ("add-scalar", lambda x: x + "q", ["bcq", None, "dfq"]),
    ("append-scalar", lambda x: x.append("qq"), ["bcqq", None, "dfqq"]),
    ("prepend-scalar", lambda x: x.prepend("qq"), ["qqbc", None, "qqdf"]),
    ("split-scalar", lambda x: x.split("c"), [["b", ""], None, ["df"]]),
    (
        "partition-scalar",
        lambda x: x.partition("c"),
        [["b", "c", ""], None, ["df", "", ""]],
    ),
    ("startswith-scalar", lambda x: x.startswith("b"), [True, None, False]),
    ("endswith-scalar", lambda x: x.endswith("f"), [False, None, True]),
    ("replace-scalar", lambda x: x.replace("c", "q"), ["bq", None, "df"]),
    ("nest-list", lambda x: list_.make_list(a=x), [["bc"], [None], ["df"]]),
    ("nest-dict", lambda x: dict_(a=x), [{"a": "bc"}, {"a": None}, {"a": "df"}]),
]


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    string_ops_nullable_test_cases,
)
def test_arrow_vectorizer_string_nullable_scalar_ops_tagged(
    name, weave_func, expected_output
):
    expected_value_type = weave.type_of(expected_output).object_type

    list = ["bc", None, "df"]
    for i, elem in enumerate(list):
        taggable = box.box(elem)
        list[i] = tag_store.add_tags(taggable, {"mytag": f"test{i + 1}"})

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("mytag", types.String())

    awl = arrow.to_arrow(list)
    l = weave.save(awl)
    fn = weave_internal.define_fn({"x": awl.object_type}, weave_func).val
    vec_fn = arrow.vectorize(fn)

    print("VEC", vec_fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    print("CALLED TYPE", called.type)
    assert weave.use(called).to_pylist_notags() == expected_output

    item_node = arrow.ops.index(called, 0)

    def make_exp_tag(t: types.Type):
        return tagged_value_type.TaggedValueType(
            types.TypedDict({"mytag": types.String()}),
            t,
        )

    expected_type_obj_type = make_exp_tag(expected_value_type)
    # The general test here is not general enough to check these properly.
    # Specifcially, the general test assumes the vecotrize lambdas contians all
    # tag flow ops. However, list and dict constructors are explicitly (and
    # correctly) not tag flow ops and therefore the tags are only present on the
    # elements of the list and dict.
    if name == "nest-dict":
        item_node = item_node["a"]
        expected_type_obj_type = types.TypedDict(
            {
                **expected_value_type.property_types,
                "a": make_exp_tag(
                    types.optional(expected_value_type.property_types["a"])
                ),
            }
        )
    elif name == "nest-list":
        item_node = item_node[0]
        expected_type_obj_type = types.List(
            make_exp_tag(types.optional(expected_value_type.object_type))
        )

    # check that tags are propagated
    assert weave.use(tag_getter_op(item_node)) == "test1"

    # NOTE: This optionality is needed because some arrow ops eagerly declare
    # optional returns. See number.py and string.py for commentary on the subject.
    assert arrow.ArrowWeaveListType(types.optional(expected_type_obj_type)).assign_type(
        called.type
    )


@pytest.mark.parametrize(
    "name,input_datal,input_datar,weave_func,expected_output",
    [
        (
            "merge",
            [{"b": "c"}, None, {"b": "q"}],
            [{"c": 4}, {"c": 5}, {"c": 6}],
            lambda x, y: x.merge(y),
            [{"b": "c", "c": 4}, None, {"b": "q", "c": 6}],
        ),
        (
            "merge-overwrite",
            [{"b": "c"}, None, {"b": "q"}],
            [{"b": "g"}, {"b": "q"}, {"b": "a"}],
            lambda x, y: x.merge(y),
            [{"b": "g"}, None, {"b": "a"}],
        ),
        (
            "merge-dicts",
            [{"b": "c"}, None, {"b": "q"}],
            [{"c": {"a": 2}}, {"c": {"a": 3}}, {"c": {"a": 4}}],
            lambda x, y: x.merge(y),
            [
                {"b": "c", "c": {"a": 2}},
                None,
                {"b": "q", "c": {"a": 4}},
            ],
        ),
    ],
)
def test_arrow_typeddict_nullable_merge(
    input_datal, input_datar, name, weave_func, expected_output
):
    l = weave.save(arrow.to_arrow(input_datal))
    r = weave.save(arrow.to_arrow(input_datar))

    fn = weave_internal.define_fn(
        {
            "x": weave.type_of(input_datal).object_type,
            "y": weave.type_of(input_datar).object_type,
        },
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l, "y": r})
    awl = weave.use(called)
    assert awl.to_pylist_raw() == expected_output
    assert called.type == arrow.ArrowWeaveListType(
        weave.type_of(expected_output).object_type
    )
    assert awl.object_type == weave.type_of(expected_output).object_type


@pytest.mark.parametrize(
    "name,input_data_vec,weave_func,expected_output",
    [
        (
            "merge-scalar-vec",
            [{"b": "c"}, {"b": "q"}],
            lambda x: dispatch.RuntimeConstNode(
                types.TypedDict({"c": types.Int()}), {"c": 4}
            ).merge(x),
            [{"b": "c", "c": 4}, {"b": "q", "c": 4}],
        ),
        (
            "merge-vec-scalar",
            [{"b": "c"}, None, {"b": "q"}],
            lambda x: x.merge({"c": 4}),
            [{"b": "c", "c": 4}, None, {"b": "q", "c": 4}],
        ),
    ],
)
def test_arrow_typeddict_nullable_scalar_vector_merge(
    input_data_vec, name, weave_func, expected_output
):
    data = weave.save(arrow.to_arrow(input_data_vec))

    fn = weave_internal.define_fn(
        {
            "x": weave.type_of(input_data_vec).object_type,
        },
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn, strict=True)
    called = weave_internal.call_fn(vec_fn, {"x": data})
    awl = weave.use(called)
    assert awl.to_pylist_raw() == expected_output
    assert called.type == arrow.ArrowWeaveListType(
        weave.type_of(expected_output).object_type
    )
    assert awl.object_type == weave.type_of(expected_output).object_type


nullable_pick_cases = [
    (
        "pick",
        [None, {"a": 2.0, "b": "G"}, {"a": 3.0, "b": "q"}],
        lambda x: x.pick("a"),
        [None, 2.0, 3.0],
    ),
    (
        "pick-nested",
        [
            {"a": None, "c": "d"},
            {"a": {"b": 2.0}, "c": "G"},
            {"a": {"b": 3.0}, "c": "q"},
        ],
        lambda x: x.pick("a").pick("b"),
        [None, 2.0, 3.0],
    ),
]


@pytest.mark.parametrize(
    "name,input_data,weave_func,expected_output", nullable_pick_cases
)
def test_arrow_typeddict_pick_nullable(input_data, name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow(input_data))
    fn = weave_internal.define_fn(
        {"x": weave.type_of(input_data).object_type}, weave_func
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_raw() == expected_output
    assert called.type == arrow.ArrowWeaveListType(
        weave.type_of(expected_output).object_type
    )


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    [
        ("floor", lambda x: Number.floor(x), [1.0, 2.0, 3.0]),
        ("ceil", lambda x: Number.ceil(x), [2.0, 3.0, 4.0]),
    ],
)
def test_arrow_floor_ceil_vectorized(name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow([1.1, 2.5, 3.9]))
    fn = weave_internal.define_fn({"x": weave.types.Float()}, weave_func).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_raw() == expected_output


pick_cases = [
    (
        "pick",
        [{"a": 1.0, "b": "c"}, {"a": 2.0, "b": "G"}, {"a": 3.0, "b": "q"}],
        lambda x: x.pick("a"),
        [1.0, 2.0, 3.0],
    ),
    (
        "pick-nested",
        [
            {"a": {"b": 1.0}, "c": "d"},
            {"a": {"b": 2.0}, "c": "G"},
            {"a": {"b": 3.0}, "c": "q"},
        ],
        lambda x: x.pick("a").pick("b"),
        [1.0, 2.0, 3.0],
    ),
]


@pytest.mark.parametrize("name,input_data,weave_func,expected_output", pick_cases)
def test_arrow_typeddict_pick(input_data, name, weave_func, expected_output):
    l = weave.save(arrow.to_arrow(input_data))
    fn = weave_internal.define_fn(
        {"x": weave.type_of(input_data).object_type}, weave_func
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_raw() == expected_output
    assert called.type == arrow.ArrowWeaveListType(
        weave.type_of(expected_output).object_type
    )


@pytest.mark.parametrize("name,input_data,weave_func,expected_output", pick_cases)
def test_arrow_typeddict_pick_tagged(input_data, name, weave_func, expected_output):
    # tag each dict and one of its values
    for i, elem in enumerate(input_data):
        taggable = box.box(elem)
        input_data[i] = tag_store.add_tags(
            taggable, {"dict_tag": f"{input_data}[{i}] = {elem}"}
        )
        input_data[i]["a"] = tag_store.add_tags(
            box.box(input_data[i]["a"]),
            {"first_level_tag": f"{input_data}[{i}]['a'] = {elem['a']}"},
        )
        if name == "pick-nested":
            input_data[i]["a"]["b"] = tag_store.add_tags(
                box.box(input_data[i]["a"]["b"]),
                {"second_level_tag": f"{input_data}[{i}]['a']['b'] = {elem['a']['b']}"},
            )

    l = weave.save(arrow.to_arrow(input_data))
    fn = weave_internal.define_fn(
        {"x": weave.type_of(input_data).object_type}, weave_func
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})

    assert weave.use(called).to_pylist_notags() == expected_output
    expected_output_object_type = tagged_value_type.TaggedValueType(
        types.TypedDict(
            property_types={
                "dict_tag": types.String(),
                "first_level_tag": types.String(),
                **(
                    {"second_level_tag": types.String()}
                    if name == "pick-nested"
                    else {}
                ),
            }
        ),
        types.Float(),
    )
    assert called.type == arrow.ArrowWeaveListType(expected_output_object_type)

    tag_getter_op = make_tag_getter_op.make_tag_getter_op(
        "first_level_tag", types.String()
    )

    for i, elem in enumerate(input_data):
        tag = f"{input_data}[{i}]['a'] = {elem['a']}"
        assert weave.use(tag_getter_op(arrow.ops.index(called, i))) == tag


@pytest.mark.parametrize(
    "name,input_datal,input_datar,weave_func,expected_output",
    [
        (
            "merge",
            [{"b": "c"}, {"b": "G"}, {"b": "q"}],
            [{"c": 4}, {"c": 5}, {"c": 6}],
            lambda x, y: x.merge(y),
            [{"b": "c", "c": 4}, {"b": "G", "c": 5}, {"b": "q", "c": 6}],
        ),
        (
            "merge-overwrite",
            [{"b": "c"}, {"b": "G"}, {"b": "q"}],
            [{"b": "g"}, {"b": "q"}, {"b": "a"}],
            lambda x, y: x.merge(y),
            [{"b": "g"}, {"b": "q"}, {"b": "a"}],
        ),
        (
            "merge-dicts",
            [{"b": "c"}, {"b": "G"}, {"b": "q"}],
            [{"c": {"a": 2}}, {"c": {"a": 3}}, {"c": {"a": 4}}],
            lambda x, y: x.merge(y),
            [
                {"b": "c", "c": {"a": 2}},
                {"b": "G", "c": {"a": 3}},
                {"b": "q", "c": {"a": 4}},
            ],
        ),
    ],
)
def test_arrow_typeddict_merge(
    input_datal, input_datar, name, weave_func, expected_output
):
    l = weave.save(arrow.to_arrow(input_datal))
    r = weave.save(arrow.to_arrow(input_datar))
    fn = weave_internal.define_fn(
        {
            "x": weave.type_of(input_datal).object_type,
            "y": weave.type_of(input_datar).object_type,
        },
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l, "y": r})
    awl = weave.use(called)
    assert awl.to_pylist_raw() == expected_output
    assert called.type == arrow.ArrowWeaveListType(
        weave.type_of(expected_output).object_type
    )
    assert awl.object_type == weave.type_of(expected_output).object_type


def test_arrow_typeddict_simple_merge_tagged():
    input_datal = [{"b": "c"}, {"b": "G"}, {"b": "q"}]
    input_datar = [{"c": 4}, {"c": 5}, {"c": 6}]
    weave_func = lambda x, y: x.merge(y)
    expected_output = [{"b": "c", "c": 4}, {"b": "G", "c": 5}, {"b": "q", "c": 6}]

    for i, elem in enumerate(input_datal):
        input_datal[i]["b"] = tag_store.add_tags(box.box(elem["b"]), {"tag": "a"})

    l = weave.save(arrow.to_arrow(input_datal))
    r = weave.save(arrow.to_arrow(input_datar))
    fn = weave_internal.define_fn(
        {
            "x": types.TypedDict({"b": types.String()}),
            "y": types.TypedDict({"c": types.Int()}),
        },
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn)
    # here we call refine graph so we can have called be a RuntimeOutputNode,
    # which allows the dispatch we us in the tag getter op assert
    called = weave_internal.refine_graph(
        weave_internal.call_fn(vec_fn, {"x": l, "y": r})
    )
    awl = weave.use(called)
    assert awl.to_pylist_notags() == expected_output

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("tag", types.String())
    tag = weave.use(tag_getter_op(called[0]["b"]))
    assert tag == "a"

    assert called.type == arrow.ArrowWeaveListType(
        types.TypedDict(
            {
                "b": tagged_value_type.TaggedValueType(
                    types.TypedDict({"tag": types.String()}), types.String()
                ),
                "c": types.Int(),
            }
        )
    )


def test_arrow_typeddict_overwrite_merge_tagged():
    input_datal = [{"b": "c"}, {"b": "G"}, {"b": "q"}]
    input_datar = [{"b": "g"}, {"b": "q"}, {"b": "a"}]
    weave_func = lambda x, y: x.merge(y)
    expected_output = [{"b": "g"}, {"b": "q"}, {"b": "a"}]

    for i, elem in enumerate(input_datar):
        input_datar[i]["b"] = tag_store.add_tags(box.box(elem["b"]), {"tag": "a"})

    l = weave.save(arrow.to_arrow(input_datal))
    r = weave.save(arrow.to_arrow(input_datar))
    fn = weave_internal.define_fn(
        {
            "x": types.TypedDict({"b": types.String()}),
            "y": types.TypedDict({"b": types.String()}),
        },
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn)
    # here we call refine graph so we can have called be a RuntimeOutputNode,
    # which allows the dispatch we us in the tag getter op assert
    called = weave_internal.refine_graph(
        weave_internal.call_fn(vec_fn, {"x": l, "y": r})
    )
    awl = weave.use(called)
    assert awl.to_pylist_notags() == expected_output

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("tag", types.String())
    tag = weave.use(tag_getter_op(called[0]["b"]))
    assert tag == "a"

    assert called.type == arrow.ArrowWeaveListType(
        types.TypedDict(
            {
                "b": tagged_value_type.TaggedValueType(
                    types.TypedDict({"tag": types.String()}), types.String()
                ),
            }
        )
    )


def test_arrow_typeddict_dicts_merge_tagged():
    input_datal = [{"b": "c"}, {"b": "G"}, {"b": "q"}]
    input_datar = [{"c": {"a": 2}}, {"c": {"a": 3}}, {"c": {"a": 4}}]
    weave_func = lambda x, y: x.merge(y)
    expected_output = [
        {"b": "c", "c": {"a": 2}},
        {"b": "G", "c": {"a": 3}},
        {"b": "q", "c": {"a": 4}},
    ]

    for i, elem in enumerate(input_datar):
        input_datar[i]["c"]["a"] = tag_store.add_tags(
            box.box(elem["c"]["a"]), {"tag": "a"}
        )

    l = weave.save(arrow.to_arrow(input_datal))
    r = weave.save(arrow.to_arrow(input_datar))
    fn = weave_internal.define_fn(
        {
            "x": types.TypedDict({"b": types.String()}),
            "y": types.TypedDict({"c": types.TypedDict({"a": types.Int()})}),
        },
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn)
    # here we call refine graph so we can have called be a RuntimeOutputNode,
    # which allows the dispatch we us in the tag getter op assert
    called = weave_internal.refine_graph(
        weave_internal.call_fn(vec_fn, {"x": l, "y": r})
    )
    awl = weave.use(called)
    assert awl.to_pylist_notags() == expected_output

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("tag", types.String())
    tag = weave.use(tag_getter_op(called[0]["c"]["a"]))
    assert tag == "a"

    assert called.type == arrow.ArrowWeaveListType(
        types.TypedDict(
            {
                "b": types.String(),
                "c": types.TypedDict(
                    {
                        "a": tagged_value_type.TaggedValueType(
                            types.TypedDict({"tag": types.String()}), types.Int()
                        )
                    }
                ),
            }
        )
    )


def test_arrow_typeddict_nested_merge_tagged():
    """Tests that nested merging is disabled."""
    input_datal = [{"a": {"c": 1}}, {"a": {"c": 2}}, {"a": {"c": 3}}]
    input_datar = [{"a": {"b": 2}}, {"a": {"b": 4}}, {"a": {"b": 6}}]
    weave_func = lambda x, y: x.merge(y)
    expected_output = [
        {"a": {"b": 2}},
        {"a": {"b": 4}},
        {"a": {"b": 6}},
    ]

    for i, elem in enumerate(input_datal):
        input_datal[i]["a"]["c"] = tag_store.add_tags(
            box.box(elem["a"]["c"]), {"tag": "a"}
        )

    l = weave.save(arrow.to_arrow(input_datal))
    r = weave.save(arrow.to_arrow(input_datar))
    fn = weave_internal.define_fn(
        {
            "x": weave.type_of(input_datal).object_type,
            "y": weave.type_of(input_datar).object_type,
        },
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l, "y": r})
    awl = weave.use(called)
    assert awl.to_pylist_notags() == expected_output

    assert called.type == arrow.ArrowWeaveListType(
        types.TypedDict(
            {
                "a": types.TypedDict(
                    {
                        "b": types.Int(),
                    }
                )
            }
        )
    )


def test_arrow_dict_tagged():
    to_tag = box.box([1, 2, 3])
    for i, elem in enumerate(to_tag):
        taggable = box.box(elem)
        to_tag[i] = tag_store.add_tags(taggable, {"a": f"{elem}"})
    tag_store.add_tags(to_tag, {"outer": "tag"})
    a = weave.save(arrow.to_arrow(to_tag))
    b = weave.save(arrow.to_arrow(["a", "b", "c"]))
    expected_output = [{"a": 1, "b": "a"}, {"a": 2, "b": "b"}, {"a": 3, "b": "c"}]
    weave_func = lambda a, b: ops.dict_(a=a, b=b)
    fn = weave_internal.define_fn(
        {"a": a.type.object_type, "b": b.type.object_type},
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"a": a, "b": b})

    # this is needed becasue the argtypes of fn do not include the tag types on a and b. so when we call
    # vec fn on a and b, the type of called does not include the tags on a and b. refining the graph
    # causes the type system to look at the types of a and b and recompute the type of called accordingly.
    called = weave_internal.refine_graph(called)
    awl = weave.use(called)
    assert awl.to_pylist_notags() == expected_output

    # tags should not flow to output list because the op has no named arguments,
    # just varargs.
    assert not isinstance(weave.type_of(awl), tagged_value_type.TaggedValueType)

    assert called.type == arrow.ArrowWeaveListType(
        types.TypedDict(
            {
                "a": tagged_value_type.TaggedValueType(
                    types.TypedDict({"outer": types.String(), "a": types.String()}),
                    types.Int(),
                ),
                "b": types.String(),
            }
        )
    )

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("a", types.String())
    tag_node = tag_getter_op(called[0]["a"])
    assert weave.use(tag_node) == "1"

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("outer", types.String())
    tag_node = tag_getter_op(called[0]["a"])
    assert weave.use(tag_node) == "tag"


def test_arrow_dict():
    a = weave.save(arrow.to_arrow([1, 2, 3]))
    b = weave.save(arrow.to_arrow(["a", "b", "c"]))
    expected_output = [{"a": 1, "b": "a"}, {"a": 2, "b": "b"}, {"a": 3, "b": "c"}]
    weave_func = lambda a, b: ops.dict_(a=a, b=b)
    fn = weave_internal.define_fn(
        {"a": weave.types.Int(), "b": weave.types.String()},
        weave_func,
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"a": a, "b": b})
    awl = weave.use(called)
    assert awl.to_pylist_raw() == expected_output
    assert called.type == arrow.ArrowWeaveListType(
        weave.type_of(expected_output).object_type
    )
    assert awl.object_type == weave.type_of(expected_output).object_type


# We have disabled auto-vectorization for all custom ops. Changing this test so that
# it creates a built-in op instead also doesn't work. In that we automatically create
# a mapped op. vectorizes determines that it can call the op (via the mapped op) in
# "Part 1", so it doesn't try to weavify and vectorize. Since we're not relying on
# auto-vectorization now, this test is disabled.
@pytest.mark.skip("auto vectorize disabled for custom ops")
def test_vectorize_works_recursively_on_weavifiable_op():
    # this op is weavifiable because it just calls add
    @weave.op()
    def add_one124815(x: int) -> int:
        return x + 1

    weave_fn = weave_internal.define_fn(
        {"x": weave.types.Int()}, lambda x: add_one124815(x)
    ).val
    vectorized = arrow.vectorize(weave_fn)
    expected = vectorized.to_json()
    print("test_vectorize_works_recursively_on_weavifiable_op.expected", expected)
    assert expected == {
        "nodeType": "output",
        "type": {
            "type": "ArrowWeaveList",
            "_base_type": {"type": "list"},
            "objectType": "int",
        },
        "fromOp": {
            "name": "ArrowWeaveListNumber-add",
            "inputs": {
                "self": {
                    "nodeType": "var",
                    "type": {
                        "type": "ArrowWeaveList",
                        "_base_type": {"type": "list"},
                        "objectType": "int",
                    },
                    "varName": "x",
                },
                "other": {"nodeType": "const", "type": "int", "val": 1},
            },
        },
    }


def test_vectorize_inner_lambdas():
    l = weave.save(arrow.to_arrow([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
    inner_fn = weave_internal.define_fn(
        {"row": l.type.object_type.object_type}, lambda row: row + 1
    )
    fn = weave_internal.define_fn(
        {"row": l.type.object_type}, lambda row: row.map(inner_fn)
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": l})
    assert list(weave.use(called)) == [[2, 3, 4], [5, 6, 7], [8, 9, 10]]


def test_unboxing_in_broadcasting():
    col1 = weave.save(arrow.to_arrow([1, 2, 3]))
    col2 = box.box(True)
    col2 = tag_store.add_tags(col2, {"col2_tag": "col2_val"})
    col2 = weave.save(col2)

    fn_body = lambda v: ops.dict_(col1=v, col2=col2)
    fn = weave_internal.define_fn({"v": types.Int()}, fn_body).val
    vectorized_fn = arrow.vectorize(fn)
    called_fn = weave_internal.call_fn(vectorized_fn, {"v": col1})
    result = weave.use(called_fn)

    answer = [
        {"col1": 1, "col2": True},
        {"col1": 2, "col2": True},
        {"col1": 3, "col2": True},
    ]

    for i in range(3):
        assert result._index(i) == answer[i]

    tag_getter = make_tag_getter_op.make_tag_getter_op("col2_tag")
    tag_getter_result = weave.use(tag_getter(arrow.ops.index(called_fn, 0)["col2"]))
    assert tag_getter_result == "col2_val"


@pytest.mark.parametrize(
    "name,weave_func,expected_output",
    [
        ("bool-and-scalar", lambda x: Boolean.bool_and(x, True), [False, None, True]),
        ("bool-or-scalar", lambda x: Boolean.bool_or(x, True), [True, None, True]),
        ("bool-not", lambda x: Boolean.bool_not(x), [True, None, False]),
    ],
)
def test_vectorized_bool_ops(name, weave_func, expected_output):
    expected_value_type = weave.type_of(expected_output).object_type

    list = [False, None, True]
    for i, elem in enumerate(list):
        taggable = box.box(elem)
        list[i] = tag_store.add_tags(taggable, {"mytag": f"test{i + 1}"})

    tag_getter_op = make_tag_getter_op.make_tag_getter_op("mytag", types.String())

    awl = arrow.to_arrow(list)
    l = weave.save(awl)
    fn = weave_internal.define_fn({"x": awl.object_type}, weave_func).val
    vec_fn = arrow.vectorize(fn)

    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_notags() == expected_output

    item_node = arrow.ops.index(called, 0)

    def make_exp_tag(t: types.Type):
        return tagged_value_type.TaggedValueType(
            types.TypedDict({"mytag": types.String()}),
            t,
        )

    expected_type_obj_type = make_exp_tag(expected_value_type)
    # check that tags are propagated
    assert weave.use(tag_getter_op(item_node)) == "test1"

    # NOTE: This optionality is needed because some arrow ops eagerly declare
    # optional returns. See number.py and string.py for commentary on the subject.
    assert arrow.ArrowWeaveListType(types.optional(expected_type_obj_type)).assign_type(
        called.type
    )


def test_vectorize_index_simple():
    l = weave.save(arrow.to_arrow([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
    fn = weave_internal.define_fn({"row": l.type.object_type}, lambda row: row[0]).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": l})
    assert list(weave.use(called)) == [1, 4, 7]


def test_vectorize_index_out_of_bounds():
    l = weave.save(arrow.to_arrow([[1, 2, 3], [1, 2], [4, 5, 6], [], [7, 8, 9]]))
    fn = weave_internal.define_fn({"row": l.type.object_type}, lambda row: row[2]).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": l})
    assert list(weave.use(called)) == [3, None, 6, None, 9]


def test_vectorize_index_nested():
    l = weave.save(arrow.to_arrow([[[1, 2, 3]], [[4, 5, 6]], [[7, 8, 9]]]))
    fn = weave_internal.define_fn(
        {"row": l.type.object_type}, lambda row: row[0][0]
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": l})
    assert list(weave.use(called)) == [1, 4, 7]


def test_vectorize_index_dependent():
    l = weave.save(arrow.to_arrow([[1, 1, 2, 3], [2, 4, 5, 6], [3, 7, 8, 9]]))
    fn = weave_internal.define_fn(
        {"row": l.type.object_type}, lambda row: row[row[0]]
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": l})
    assert list(weave.use(called)) == [1, 5, 9]


def test_vectorize_special_pick():
    l = weave.save(arrow.to_arrow(["a", None, "b", None, "c"]))
    dict_node = weave_internal.make_const_node(
        types.Dict(types.String(), types.String()),
        {
            "a": 1,
            "b": 2,
            "c": 3,
            "d": 4,
        },
    )
    fn = weave_internal.define_fn(
        {"row": l.type.object_type}, lambda row: dict_node[row]
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": l})
    assert list(weave.use(called)) == [1, None, 2, None, 3]


def test_vectorize_function_with_const_first_arg():
    string = "hello"
    data = ["hello", "world", "hello", "world", "hello", "world"]
    awl_node = weave.save(arrow.to_arrow(data))
    string_node = weave_internal.make_const_node(types.String(), string)
    fn = weave_internal.define_fn(
        {"row": types.String()}, lambda row: string_node == row
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": awl_node})
    assert weave.use(called).to_pylist_raw() == [True, False, True, False, True, False]


def test_isnone():
    a = [1, None, False, 0, "", "a"]
    awl = arrow.to_arrow(a)
    l = weave.save(awl)
    fn = weave_internal.define_fn({"x": awl.object_type}, lambda x: x.isNone()).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_notags() == [
        False,
        True,
        False,
        False,
        False,
        False,
    ]


@pytest.mark.parametrize(
    "weave_func, expected",
    [
        (lambda x: Number.__eq__(5, x), [False, False, False, False, True]),
        (lambda x: Number.__ne__(5, x), [True, True, True, True, False]),
    ],
)
def test_right_equals_not_equals_elements_tagged(weave_func, expected):
    # make an arrow weave list with tagged elements
    data = [1, 2, 3, 4, 5]
    for i in range(len(data)):
        data[i] = tag_store.add_tags(box.box(data[i]), {"mytag": "test" + str(i)})
    awl = arrow.to_arrow(data)
    l = weave.save(awl)
    fn = weave_internal.define_fn({"x": awl.object_type}, weave_func).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_notags() == expected


def test_listnumbersmax_with_tagged_inner_elements():
    data = [[1, None, 3], [4, -1, 3], [4, 5, 6]]
    data = [[{"TMax": elem} for elem in row] for row in data]
    for i in range(len(data)):
        for j in range(len(data[i])):
            data[i][j] = tag_store.add_tags(
                box.box(data[i][j]), {"mytag": "test" + str(i) + str(j)}
            )
    awl = arrow.to_arrow(data)
    l = weave.save(awl)

    mapped_list_node = list_.List._listmap(l, lambda row: row["TMax"])

    assert mapped_list_node.type == types.List(
        types.List(
            tagged_value_type.TaggedValueType(
                types.TypedDict({"mytag": types.String()}), types.optional(types.Int())
            )
        )
    )

    res_node = arraylist_ops.list_numbers_max(
        convert_ops.list_to_arrow(mapped_list_node)
    )

    assert weave.use(res_node).to_pylist_tagged() == [3, 4, 6]


def test_list_numbers_equal_notequal():
    list_array1 = arrow.to_arrow([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    list_array2 = arrow.to_arrow([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    list_array3 = arrow.to_arrow([[1, 2], [4, 5, 6], [7, 8, 9]])
    list_array4 = arrow.to_arrow([[1, 2, 3], None, [7, 8, 9]])
    list_array5 = arrow.to_arrow([[1, 2], None, [7, 8, 9]])
    list_array6 = arrow.to_arrow([[1, 2, 3], [None, 5, 6], [7, 8, 9]])
    list_array7 = arrow.to_arrow([[1, 2], [None, 5, 6], [7, 8, 9]])

    l1 = weave.save(list_array1)
    l2 = weave.save(list_array2)
    l3 = weave.save(list_array3)
    l4 = weave.save(list_array4)
    l5 = weave.save(list_array5)
    l6 = weave.save(list_array6)
    l7 = weave.save(list_array7)

    assert weave.use(l1 == l2).to_pylist_notags() == [True, True, True]
    assert weave.use(l1 == l3).to_pylist_notags() == [False, True, True]
    assert weave.use(l4 == l2).to_pylist_notags() == [True, False, True]
    assert weave.use(l4 == l5).to_pylist_notags() == [False, True, True]
    assert weave.use(l6 == l2).to_pylist_notags() == [True, False, True]
    assert weave.use(l6 == l7).to_pylist_notags() == [False, True, True]


def test_vectorized_prop_op_gql_pick():
    runs = [
        wdt.Run({"id": "A", "key2": 1}),
        wdt.Run({"id": "B", "key2": 1}),
        wdt.Run({"id": "C", "key2": 1}),
    ]
    for run in runs:
        tag_store.add_tags(run, {"mytag": "test" + run["id"]})
    awl = arrow.to_arrow(runs)
    l = weave.save(awl)

    fn = weave_internal.define_fn(
        {"x": awl.object_type}, lambda x: run_ops.run_id(x)
    ).val
    vec_fn = arrow.vectorize(fn, strict=True)
    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_notags() == ["A", "B", "C"]
    assert weave.use(called).to_pylist_raw() == [
        {"_tag": {"mytag": "testA"}, "_value": "A"},
        {"_tag": {"mytag": "testB"}, "_value": "B"},
        {"_tag": {"mytag": "testC"}, "_value": "C"},
    ]


def test_vectorize_run_runtime():
    runs = [
        wdt.Run({"id": "A", "computeSeconds": 1}),
        wdt.Run({"id": "B", "computeSeconds": 2}),
        wdt.Run({"id": "C", "computeSeconds": 3}),
    ]

    awl = arrow.to_arrow(runs)
    l = weave.save(awl)

    fn = weave_internal.define_fn(
        {"x": awl.object_type}, lambda x: run_ops.runtime(x)
    ).val

    vec_fn = arrow.vectorize(fn, strict=True)

    called = weave_internal.call_fn(vec_fn, {"x": l})
    assert weave.use(called).to_pylist_notags() == [1, 2, 3]

    # it finds an AWL op
    assert "ArrowWeaveList" in vec_fn.from_op.name
    assert "mapped" not in vec_fn.from_op.name


def test_boxed_null_in_array_equal():
    lhs = pa.array([1, None, 3])
    rhs = box.box(None)
    assert util.equal(lhs, rhs).to_pylist() == [False, True, False]
    assert util.not_equal(lhs, rhs).to_pylist() == [True, False, True]


def test_duration():
    dt1 = datetime.timedelta(seconds=1)
    dt2 = datetime.timedelta(seconds=2)
    dt3 = datetime.timedelta(seconds=3)

    awl_node = weave.save(arrow.to_arrow([dt1, dt2, dt3]))

    fn = weave_internal.define_fn(
        {"row": awl_node.type.object_type},
        lambda row: date.timedelta_total_seconds(row),
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": awl_node})
    assert list(weave.use(called)) == [1, 2, 3]


def test_date_add():
    dt = datetime.datetime.now()

    # TODO: remove ms truncation when we can store more precise datetimes in arrow
    dt = dt.replace(microsecond=dt.microsecond // 1000 * 1000)
    dt1 = dt + datetime.timedelta(days=1)

    awl_node = weave.save(arrow.to_arrow([dt, dt1]))
    fn = weave_internal.define_fn(
        {"row": awl_node.type.object_type},
        lambda row: date.datetime_add(row, datetime.timedelta(days=1)),
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": awl_node})
    actual = list(weave.use(called))

    # TODO: remove timezone addition when we have a better way of handling timezones in arrow
    dt1utc = dt1.astimezone(datetime.timezone.utc)
    expected = [dt1utc, dt1utc + datetime.timedelta(days=1)]
    assert actual == expected


def test_date_sub():
    dt = datetime.datetime.now(tz=datetime.timezone.utc)

    # TODO: remove ms truncation when we can store more precise datetimes in arrow
    dt = dt.replace(microsecond=dt.microsecond // 1000 * 1000)
    dt1 = dt + datetime.timedelta(days=1)

    awl_node = weave.save(arrow.to_arrow([dt, dt1]))
    fn = weave_internal.define_fn(
        {"row": awl_node.type.object_type},
        lambda row: date.datetime_sub(row, dt),
    ).val
    vec_fn = arrow.vectorize(fn)
    called = weave_internal.call_fn(vec_fn, {"row": awl_node})
    actual = list(weave.use(called))

    expected = [datetime.timedelta(0), datetime.timedelta(days=1)]
    assert actual == expected



import json

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import gql_json_cache, mappers
from weave.legacy.weave.mappers_weave import UnionMapper
from weave.legacy.weave.partial_object import PartialObjectType

from .mappers_python_def import (
    BoolToPyBool,
    DictToPyDict,
    IntToPyInt,
    ListToPyList,
    NoneToPyNone,
    ObjectDictToObject,
    PyFloatToFloat,
    TypedDictToPyDict,
)


class GQLTypedDictToTypedDict(TypedDictToPyDict):
    def apply(self, obj):
        # Hack!! Our output types throughout GQL are all wrong. They are non-nullable
        # when in reality GQL responses can frequently be optional.
        if obj is None:
            return None
        return super().apply(obj)


class GQLUnionToUnion(UnionMapper):
    def apply(self, obj):
        if obj is None:
            return None
        if self.is_single_object_nullable:
            non_null_mapper = next(
                filter(
                    lambda m: not types.NoneType().assign_type(m.type),
                    self._member_mappers,
                )
            )
            return non_null_mapper.apply(obj)
        elif "__typename" in obj:
            typename = obj["__typename"]
            for mapper in self._member_mappers:
                if isinstance(mapper.type, types.TypedDict):
                    typename_type = mapper.type.property_types.get("__typename", None)
                    if typename_type is not None:
                        assert isinstance(typename_type, types.Const)
                        mapper_typename = typename_type.val
                        if mapper_typename == typename:
                            return mapper.apply(obj)
        raise errors.WeaveValueError(
            f"Cant find a member of union {self.type} with typename {typename}"
        )


class GQLStringToString(mappers.Mapper):
    def apply(self, obj):
        if isinstance(obj, str):
            return obj

        # the weave type of the graphql JSON type is a string, so here we
        # convert the json object to a string, but we also cache the python
        # object, in case we need to use it later in an op (like history or summary)
        serialized = json.dumps(obj)
        gql_json_cache.cache_json(serialized, obj)
        return serialized


class GQLTimestampToTimestamp(mappers.Mapper):
    def apply(self, obj):
        assert isinstance(self.type, types.Timestamp)
        return self.type.from_isostring(obj)


class GQLConstToConst(mappers.Mapper):
    def apply(self, obj):
        return obj


def map_from_gql_payload_(type, mapper, artifact, path=[], mapper_options=None):
    if isinstance(type, PartialObjectType):
        return DictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.NoneType):
        return NoneToPyNone(type, mapper, artifact, path)
    elif isinstance(type, types.Boolean):
        return BoolToPyBool(type, mapper, artifact, path)
    elif isinstance(type, types.Int):
        return IntToPyInt(type, mapper, artifact, path)
    elif isinstance(type, types.Float):
        return PyFloatToFloat(type, mapper, artifact, path)
    elif isinstance(type, types.Number):
        return PyFloatToFloat(type, mapper, artifact, path)
    elif isinstance(type, types.String):
        return GQLStringToString(type, mapper, artifact, path)
    elif isinstance(type, types.Timestamp):
        return GQLTimestampToTimestamp(type, mapper, artifact, path)
    elif isinstance(type, types.ObjectType):
        return ObjectDictToObject(type, mapper, artifact, path)
    elif isinstance(type, types.TypedDict):
        return GQLTypedDictToTypedDict(type, mapper, artifact, path)
    elif isinstance(type, types.Dict):
        return DictToPyDict(type, mapper, artifact, path)
    elif isinstance(type, types.List):
        return ListToPyList(type, mapper, artifact, path)
    elif isinstance(type, types.UnionType):
        return GQLUnionToUnion(type, mapper, artifact, path)
    elif isinstance(type, types.Const):
        return GQLConstToConst(type, mapper, artifact, path)
    raise errors.WeaveValueError(f"Unknown type {type}")


map_from_gql = mappers.make_mapper(map_from_gql_payload_)



import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave.language_features.tagging import tagged_value_type


def test_keys_type():
    a_dict = weave.save({"a": 5, "b": 6, "c": 7})
    a_dict_keys = a_dict.keys()
    assert a_dict_keys.type == weave.types.List(
        weave.types.UnionType(
            weave.types.Const(weave.types.String(), "a"),
            weave.types.Const(weave.types.String(), "b"),
            weave.types.Const(weave.types.String(), "c"),
        )
    )


def test_pick():
    obj = weave_internal.make_const_node(
        weave.types.Dict(object_type=weave.types.String()), {"a": "x"}
    )
    key = weave_internal.make_const_node(
        tagged_value_type.TaggedValueType(
            weave.types.TypedDict({"w": weave.types.String()}), weave.types.String()
        ),
        "a",
    )
    assert weave.use(weave.legacy.weave.ops.TypedDict.pick(obj, key)) == "x"


def test_pick_none_key():
    d = weave.save({"a": 5})
    assert weave.use(d[None]) == None



import typing

import pytest

import weave
from weave.legacy.weave import compile_domain, compile_table, weave_internal
from weave.legacy.weave import context_state as _context
from weave.legacy.weave.language_features.tagging import make_tag_getter_op
from weave.legacy.weave.ops_domain import run_ops

from ...legacy.weave import stitch
from ...tests import fixture_fakewandb as fwb
from . import test_wb

_loading_builtins_token = _context.set_loading_built_ins()


@weave.type()
class _TestPlanObject:
    _horse: str
    val: int

    # Because this is named "Test*", doing .horse() will tag the result
    @weave.op()
    def horse(self) -> str:
        return self._horse


@weave.type()
class _TestPlanHasObject:
    horse: str
    _obj: _TestPlanObject


@weave.op()
def dummy_no_arg_op() -> typing.List[_TestPlanObject]:
    return [_TestPlanObject("x", 1)]


# Because this is named "Test*", doing .name() will tag the result
@weave.op()
def _test_hasobj_obj(self_has_obj: _TestPlanHasObject) -> _TestPlanObject:
    return self_has_obj._obj


get_object_self_tag = make_tag_getter_op.make_tag_getter_op(
    "self", _TestPlanObject.WeaveType()
)  # type: ignore
get_hasobject_self_tag = make_tag_getter_op.make_tag_getter_op(
    "self_has_obj", _TestPlanHasObject.WeaveType()
)  # type: ignore

_context.clear_loading_built_ins(_loading_builtins_token)


def test_traverse_tags():
    obj_node = weave.save(_TestPlanObject("a", 1))
    obj_from_tag_val_node = get_object_self_tag(obj_node.horse() + "hello").val
    p = stitch.stitch([obj_from_tag_val_node])
    obj_recorder = p.get_result(obj_node)
    assert len(obj_recorder.calls) == 2
    assert obj_recorder.calls[0].node.from_op.name == "_TestPlanObject-horse"
    assert obj_recorder.calls[1].node.from_op.name == "Object-__getattr__"
    assert obj_recorder.calls[1].inputs[1].val == "val"


def test_traverse_tags_2level():
    obj_node = weave.save(_TestPlanHasObject("has", _TestPlanObject("a", 1)))
    name_add_node = obj_node._test_hasobj_obj().horse() + "hello"
    obj_from_tag_val_node = get_hasobject_self_tag(
        get_object_self_tag(name_add_node)
    ).horse
    p = stitch.stitch([obj_from_tag_val_node])
    obj_recorder = p.get_result(obj_node)
    assert len(obj_recorder.calls) == 2
    assert obj_recorder.calls[0].node.from_op.name == "op-_test_hasobj_obj"
    assert obj_recorder.calls[1].node.from_op.name == "Object-__getattr__"
    assert obj_recorder.calls[1].inputs[1].val == "horse"


def test_enter_filter():
    objs_node = weave.save([{"a": 5, "b": 6, "c": 10}, {"a": 7, "b": 8, "c": 11}])
    p = stitch.stitch([objs_node["b"], objs_node.filter(lambda obj: obj["a"] > 6)])
    obj_recorder = p.get_result(objs_node)
    calls = obj_recorder.calls
    assert len(calls) == 2
    assert calls[0].node.from_op.name == "mapped_typedDict-pick"
    assert calls[0].inputs[1].val == "b"
    assert calls[1].node.from_op.name == "typedDict-pick"
    assert calls[1].inputs[1].val == "a"


def test_lambda_using_externally_defined_node():
    objs_node = weave.save([{"a": 5, "b": 6, "c": 10}, {"a": 7, "b": 8, "c": 11}])
    # Inside the lambda, we use externally defined `objs_node`. This should
    # result in all 3 calls being recorded
    p = stitch.stitch(
        [objs_node["b"], objs_node.filter(lambda obj: obj["a"] > objs_node[0]["b"])]
    )
    obj_recorder = p.get_result(objs_node)
    calls = obj_recorder.calls
    assert len(calls) == 3
    assert calls[0].node.from_op.name == "mapped_typedDict-pick"
    assert calls[0].inputs[1].val == "b"
    assert calls[1].node.from_op.name == "typedDict-pick"
    assert calls[1].inputs[1].val == "b"
    assert calls[2].node.from_op.name == "typedDict-pick"
    assert calls[2].inputs[1].val == "a"


def test_tag_access_in_filter_expr():
    objs_node = weave.save([_TestPlanObject("a", 1), _TestPlanObject("b", 2)])
    leaf = objs_node.horse().filter(lambda obj: get_object_self_tag(obj).val > 2)
    p = stitch.stitch([leaf])
    obj_recorder = p.get_result(objs_node)
    calls = obj_recorder.calls
    assert len(calls) == 2
    assert calls[0].node.from_op.name == "mapped__TestPlanObject-horse"
    assert calls[1].node.from_op.name == "Object-__getattr__"
    assert calls[1].inputs[1].val == "val"


def test_traverse_dict():
    obj_node = weave.save(_TestPlanObject("a", 1))
    p = stitch.stitch([weave.legacy.weave.ops.dict_(x=obj_node)["x"].horse()])
    obj_recorder = p.get_result(obj_node)
    assert len(obj_recorder.calls) == 1
    assert obj_recorder.calls[0].node.from_op.name == "_TestPlanObject-horse"


def test_travese_groupby_dict():
    obj_node = weave.save([{"o": {"a": 5}, "x": 1}])
    grouped = obj_node.groupby(lambda row: weave.legacy.weave.ops.dict_(x=row["o"]))
    output = grouped[0]["x"]
    groupkey_output = grouped[0].groupkey()["x"]["a"]
    p = stitch.stitch([output, groupkey_output])
    obj_recorder = p.get_result(obj_node)

    assert compile_table.get_projection(obj_recorder) == {"o": {"a": {}}, "x": {}}


def test_zero_arg_ops():
    node = dummy_no_arg_op()
    p = stitch.stitch([node])
    obj_recorder = p.get_result(node)
    assert obj_recorder.calls == []

    p = stitch.stitch([node.horse()])
    obj_recorder = p.get_result(node)
    assert len(obj_recorder.calls) == 1
    assert obj_recorder.calls[0].node.from_op.name == "mapped__TestPlanObject-horse"

    p = stitch.stitch([node.filter(lambda x: x._get_op("horse")() != "")])
    obj_recorder = p.get_result(node)
    assert len(obj_recorder.calls) == 1
    assert obj_recorder.calls[0].node.from_op.name == "_TestPlanObject-horse"

    p = stitch.stitch([node.filter(lambda x: x._get_op("horse")() != ""), node.horse()])
    obj_recorder = p.get_result(node)
    assert len(obj_recorder.calls) == 2
    assert obj_recorder.calls[0].node.from_op.name == "_TestPlanObject-horse"
    assert obj_recorder.calls[1].node.from_op.name == "mapped__TestPlanObject-horse"


def test_shared_fn_node():
    const_list_node = weave.legacy.weave.ops.make_list(a=1, b=2)
    indexed_node = const_list_node[0]
    arr_1_node = weave.legacy.weave.ops.make_list(a=1, b=2, c=3)
    arr_2_node = weave.legacy.weave.ops.make_list(a=10, b=20, c=30)

    mapped_1_node = arr_1_node.map(
        lambda row: weave.legacy.weave.ops.dict_(item=row, const=indexed_node)
    )
    mapped_2_node = arr_2_node.map(
        lambda row: weave.legacy.weave.ops.dict_(item=row, const=indexed_node)
    )

    mapped_1_item_node = mapped_1_node["item"]
    mapped_1_const_node = mapped_1_node["const"]
    mapped_2_item_node = mapped_2_node["item"]
    mapped_2_const_node = mapped_2_node["const"]

    mapped_2_item_add_node = mapped_2_item_node + 100
    mapped_2_const_add_node = mapped_2_const_node + 100

    list_of_list_node = weave.legacy.weave.ops.make_list(
        a=mapped_1_item_node,
        b=mapped_1_const_node,
        c=mapped_2_item_add_node,
        d=mapped_2_const_add_node,
    )
    concat_node = list_of_list_node.concat()
    sum_node = concat_node.sum()

    p = stitch.stitch([sum_node])

    def assert_node_calls(node, expected_call_names):
        found_calls = set([c.node.from_op.name for c in p.get_result(node).calls])
        expected_calls = set(expected_call_names)
        assert found_calls == expected_calls

    assert_node_calls(const_list_node, ["list", "mapped_number-add"])
    assert_node_calls(indexed_node, ["list", "mapped_number-add"])
    assert_node_calls(arr_1_node, ["list"])
    assert_node_calls(arr_2_node, ["mapped_number-add"])
    assert_node_calls(mapped_1_node, [])
    assert_node_calls(mapped_2_node, [])
    assert_node_calls(mapped_1_item_node, ["list"])
    assert_node_calls(mapped_1_const_node, ["list", "mapped_number-add"])
    assert_node_calls(mapped_2_item_node, ["mapped_number-add"])
    assert_node_calls(mapped_2_const_node, ["list", "mapped_number-add"])
    assert_node_calls(mapped_2_item_add_node, ["list"])
    assert_node_calls(mapped_2_const_add_node, ["list"])
    assert_node_calls(list_of_list_node, ["concat"])
    assert_node_calls(concat_node, ["numbers-sum"])
    assert_node_calls(sum_node, [])


def test_stitch_keytypes_override_fetch_all_columns(fake_wandb):
    fake_wandb.fake_api.add_mock(test_wb.table_mock_filtered)
    keytypes_node = weave.legacy.weave.ops.object_keytypes(
        run_ops.run_tag_getter_op(
            weave.legacy.weave.ops.project("stacey", "mendeleev")
            .filteredRuns("{}", "-createdAt")
            .limit(50)
            .summary()
            .pick("table")
            .table()
            .rows()
        ).summary()
    )

    p = stitch.stitch([keytypes_node])
    object_recorder = p.get_result(keytypes_node)
    key_tree = compile_table.get_projection(object_recorder)

    # even though we have picked a specific key out of the table, we should still have an empty key tree
    # because we must fetch all columns due to keytypes
    assert key_tree == {}


def test_stitch_overlapping_tags(fake_wandb):
    fake_wandb.fake_api.add_mock(
        lambda a, b: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,
                "runs_261949318143369aa6c158af92afee03": {
                    "edges": [{"node": {**fwb.run_payload, "summaryMetrics": "{}"}}]
                },
                "runs_30ea80144a38a5c57c80d9d7f0485166": {
                    "edges": [
                        {"node": {**fwb.run_payload, "summaryMetrics": '{"a": 1}'}}
                    ]
                },
            }
        }
    )
    project_node = weave.legacy.weave.ops.project("stacey", "mendeleev")
    filtered_runs_a_node = project_node.filteredRuns("{}", "-createdAt")[0]
    summary_a_node = filtered_runs_a_node.summary()
    tagged_name_a = weave.legacy.weave.ops.run_ops.run_tag_getter_op(
        summary_a_node["a"]
    ).name()
    filtered_runs_b_node = project_node.filteredRuns("{}", "+createdAt")[0]
    summary_b_node = filtered_runs_b_node.summary()
    tagged_id_b = weave.legacy.weave.ops.run_ops.run_tag_getter_op(summary_b_node).id()

    p = stitch.stitch([tagged_name_a, tagged_id_b])

    assert len(p.get_result(project_node).tags) == 0
    assert len(p.get_result(filtered_runs_a_node).calls) == 2
    assert len(p.get_result(filtered_runs_b_node).calls) == 2


def test_refine_history_type_included_in_gql():
    project_node = weave.legacy.weave.ops.project("stacey", "mendeleev")
    runs_node = project_node.runs()
    map_node = runs_node.map(lambda row: weave.legacy.weave.ops.dict_(variant=row))
    checkpoint_node = map_node.createIndexCheckpointTag()
    index_node = checkpoint_node[0]
    pick_node = index_node["variant"]
    refine_history_node = pick_node.refine_history_type()
    sg = stitch.stitch([refine_history_node])
    assert "historyKeys" in compile_domain._get_fragment(project_node, sg)


def test_stitch_missing_key():
    a_node = weave_internal.make_const_node(weave.types.String(), "a")
    dict_node = weave.legacy.weave.ops.dict_(a=a_node)
    picked_valid = dict_node["a"] + "-suffix"
    picked_missing = dict_node["b"] + "-suffix"

    assert weave.use(picked_valid) == "a-suffix"
    assert weave.use(picked_missing) == None

    sg = stitch.stitch([picked_valid, picked_missing])

    assert len(sg.get_result(a_node).calls) == 1
